from __future__ import annotations

import io
import os
import shutil
import stat
import typing
import zipfile
from datetime import datetime

from ._base import FS
from ._errors import FileExpected, ResourceNotFound, ResourceReadOnly
from ._info import Info
from ._path import dirname, forcedir, normpath, relpath
from ._tempfs import TempFS

if typing.TYPE_CHECKING:
    from collections.abc import Collection
    from typing import IO, Any

    from ._subfs import SubFS


class ZipFS(FS):
    """Read and write zip files."""

    def __new__(
        cls, file: str | os.PathLike, write: bool = False, encoding: str = "utf-8"
    ):
        if write:
            return WriteZipFS(file, encoding)
        else:
            return ReadZipFS(file, encoding)

    if typing.TYPE_CHECKING:

        def __init__(
            self, file: str | os.PathLike, write: bool = False, encoding: str = "utf-8"
        ):
            pass


class ReadZipFS(FS):
    """A readable zip file."""

    def __init__(self, file: str | os.PathLike, encoding: str = "utf-8"):
        super().__init__()
        self._file = os.fspath(file)
        self.encoding = encoding  # unused
        self._zip = zipfile.ZipFile(file, "r")
        self._directory_fs = None

    def __repr__(self) -> str:
        return f"ReadZipFS({self._file!r})"

    def __str__(self) -> str:
        return f"<zipfs '{self._file}'>"

    def _path_to_zip_name(self, path: str) -> str:
        """Convert a path to a zip file name."""
        path = relpath(normpath(path))
        if self._directory.isdir(path):
            path = forcedir(path)
        return path

    @property
    def _directory(self) -> TempFS:
        if self._directory_fs is None:
            self._directory_fs = _fs = TempFS()
            for zip_name in self._zip.namelist():
                resource_name = zip_name
                if resource_name.endswith("/"):
                    _fs.makedirs(resource_name, recreate=True)
                else:
                    _fs.makedirs(dirname(resource_name), recreate=True)
                    _fs.create(resource_name)
        return self._directory_fs

    def close(self):
        super(ReadZipFS, self).close()
        self._zip.close()
        if self._directory_fs is not None:
            self._directory_fs.close()

    def getinfo(self, path: str, namespaces: Collection[str] | None = None) -> Info:
        namespaces = namespaces or ()
        raw_info = {}

        if path == "/":
            raw_info["basic"] = {"name": "", "is_dir": True}
            if "details" in namespaces:
                raw_info["details"] = {"type": stat.S_IFDIR}
        else:
            basic_info = self._directory.getinfo(path)
            raw_info["basic"] = {"name": basic_info.name, "is_dir": basic_info.is_dir}

            if "details" in namespaces:
                zip_name = self._path_to_zip_name(path)
                try:
                    zip_info = self._zip.getinfo(zip_name)
                except KeyError:
                    pass
                else:
                    if "details" in namespaces:
                        raw_info["details"] = {
                            "size": zip_info.file_size,
                            "type": int(
                                stat.S_IFDIR if basic_info.is_dir else stat.S_IFREG
                            ),
                            "modified": datetime(*zip_info.date_time).timestamp(),
                        }

        return Info(raw_info)

    def exists(self, path: str) -> bool:
        self.check()
        return self._directory.exists(path)

    def isdir(self, path: str) -> bool:
        self.check()
        return self._directory.isdir(path)

    def isfile(self, path: str) -> bool:
        self.check()
        return self._directory.isfile(path)

    def listdir(self, path: str) -> str:
        self.check()
        return self._directory.listdir(path)

    def makedir(self, path: str, recreate: bool = False) -> SubFS:
        self.check()
        raise ResourceReadOnly(path)

    def makedirs(self, path: str, recreate: bool = False) -> SubFS:
        self.check()
        raise ResourceReadOnly(path)

    def remove(self, path: str):
        self.check()
        raise ResourceReadOnly(path)

    def removedir(self, path: str):
        self.check()
        raise ResourceReadOnly(path)

    def removetree(self, path: str):
        self.check()
        raise ResourceReadOnly(path)

    def movedir(self, src: str, dst: str, create: bool = False):
        self.check()
        raise ResourceReadOnly(src)

    def readbytes(self, path: str) -> bytes:
        self.check()
        if not self._directory.isfile(path):
            raise ResourceNotFound(path)
        zip_name = self._path_to_zip_name(path)
        zip_bytes = self._zip.read(zip_name)
        return zip_bytes

    def open(self, path: str, mode: str = "rb", **kwargs) -> IO[Any]:
        self.check()
        if self._directory.isdir(path):
            raise FileExpected(f"{path!r} is a directory")

        zip_mode = mode[0]
        if zip_mode == "r" and not self._directory.exists(path):
            raise ResourceNotFound(f"No such file or directory: {path!r}")

        if any(m in mode for m in "wax+"):
            raise ResourceReadOnly(path)

        zip_name = self._path_to_zip_name(path)
        stream = self._zip.open(zip_name, zip_mode)
        if "b" in mode:
            if kwargs:
                raise ValueError("encoding args invalid for binary operation")
            return stream
        # Text mode
        return io.TextIOWrapper(stream, **kwargs)


class WriteZipFS(TempFS):
    """A writable zip file."""

    def __init__(self, file: str | os.PathLike, encoding: str = "utf-8"):
        super().__init__()
        self._file = os.fspath(file)
        self.encoding = encoding  # unused

    def __repr__(self) -> str:
        return f"WriteZipFS({self._file!r})"

    def __str__(self) -> str:
        return f"<zipfs-write '{self._file}'>"

    def close(self):
        base_name = os.path.splitext(self._file)[0]
        shutil.make_archive(base_name, format="zip", root_dir=self._temp_dir)
        if self._file != base_name + ".zip":
            shutil.move(base_name + ".zip", self._file)
        super().close()



from io import StringIO
import sys
import unittest

from fontTools.misc.py23 import (
    round2,
    round3,
    isclose,
    redirect_stdout,
    redirect_stderr,
)


class Round2Test(unittest.TestCase):
    """
    Test cases taken from cpython 2.7 test suite:

    https://github.com/python/cpython/blob/2.7/Lib/test/test_float.py#L748

    Excludes the test cases that are not supported when using the `decimal`
    module's `quantize` method.
    """

    def test_second_argument_type(self):
        # floats should be illegal
        self.assertRaises(TypeError, round2, 3.14159, 2.0)

    def test_halfway_cases(self):
        # Halfway cases need special attention, since the current
        # implementation has to deal with them specially.  Note that
        # 2.x rounds halfway values up (i.e., away from zero) while
        # 3.x does round-half-to-even.
        self.assertAlmostEqual(round2(0.125, 2), 0.13)
        self.assertAlmostEqual(round2(0.375, 2), 0.38)
        self.assertAlmostEqual(round2(0.625, 2), 0.63)
        self.assertAlmostEqual(round2(0.875, 2), 0.88)
        self.assertAlmostEqual(round2(-0.125, 2), -0.13)
        self.assertAlmostEqual(round2(-0.375, 2), -0.38)
        self.assertAlmostEqual(round2(-0.625, 2), -0.63)
        self.assertAlmostEqual(round2(-0.875, 2), -0.88)

        self.assertAlmostEqual(round2(0.25, 1), 0.3)
        self.assertAlmostEqual(round2(0.75, 1), 0.8)
        self.assertAlmostEqual(round2(-0.25, 1), -0.3)
        self.assertAlmostEqual(round2(-0.75, 1), -0.8)

        self.assertEqual(round2(-6.5, 0), -7.0)
        self.assertEqual(round2(-5.5, 0), -6.0)
        self.assertEqual(round2(-1.5, 0), -2.0)
        self.assertEqual(round2(-0.5, 0), -1.0)
        self.assertEqual(round2(0.5, 0), 1.0)
        self.assertEqual(round2(1.5, 0), 2.0)
        self.assertEqual(round2(2.5, 0), 3.0)
        self.assertEqual(round2(3.5, 0), 4.0)
        self.assertEqual(round2(4.5, 0), 5.0)
        self.assertEqual(round2(5.5, 0), 6.0)
        self.assertEqual(round2(6.5, 0), 7.0)

        # same but without an explicit second argument; in 3.x these
        # will give integers
        self.assertEqual(round2(-6.5), -7.0)
        self.assertEqual(round2(-5.5), -6.0)
        self.assertEqual(round2(-1.5), -2.0)
        self.assertEqual(round2(-0.5), -1.0)
        self.assertEqual(round2(0.5), 1.0)
        self.assertEqual(round2(1.5), 2.0)
        self.assertEqual(round2(2.5), 3.0)
        self.assertEqual(round2(3.5), 4.0)
        self.assertEqual(round2(4.5), 5.0)
        self.assertEqual(round2(5.5), 6.0)
        self.assertEqual(round2(6.5), 7.0)

        self.assertEqual(round2(-25.0, -1), -30.0)
        self.assertEqual(round2(-15.0, -1), -20.0)
        self.assertEqual(round2(-5.0, -1), -10.0)
        self.assertEqual(round2(5.0, -1), 10.0)
        self.assertEqual(round2(15.0, -1), 20.0)
        self.assertEqual(round2(25.0, -1), 30.0)
        self.assertEqual(round2(35.0, -1), 40.0)
        self.assertEqual(round2(45.0, -1), 50.0)
        self.assertEqual(round2(55.0, -1), 60.0)
        self.assertEqual(round2(65.0, -1), 70.0)
        self.assertEqual(round2(75.0, -1), 80.0)
        self.assertEqual(round2(85.0, -1), 90.0)
        self.assertEqual(round2(95.0, -1), 100.0)
        self.assertEqual(round2(12325.0, -1), 12330.0)
        self.assertEqual(round2(0, -1), 0.0)

        self.assertEqual(round2(350.0, -2), 400.0)
        self.assertEqual(round2(450.0, -2), 500.0)

        self.assertAlmostEqual(round2(0.5e21, -21), 1e21)
        self.assertAlmostEqual(round2(1.5e21, -21), 2e21)
        self.assertAlmostEqual(round2(2.5e21, -21), 3e21)
        self.assertAlmostEqual(round2(5.5e21, -21), 6e21)
        self.assertAlmostEqual(round2(8.5e21, -21), 9e21)

        self.assertAlmostEqual(round2(-1.5e22, -22), -2e22)
        self.assertAlmostEqual(round2(-0.5e22, -22), -1e22)
        self.assertAlmostEqual(round2(0.5e22, -22), 1e22)
        self.assertAlmostEqual(round2(1.5e22, -22), 2e22)


class Round3Test(unittest.TestCase):
    """Same as above but results adapted for Python 3 round()"""

    def test_second_argument_type(self):
        # floats should be illegal
        self.assertRaises(TypeError, round3, 3.14159, 2.0)

        # None should be allowed
        self.assertEqual(round3(1.0, None), 1)
        # the following would raise an error with the built-in Python3.5 round:
        # TypeError: 'NoneType' object cannot be interpreted as an integer
        self.assertEqual(round3(1, None), 1)

    def test_halfway_cases(self):
        self.assertAlmostEqual(round3(0.125, 2), 0.12)
        self.assertAlmostEqual(round3(0.375, 2), 0.38)
        self.assertAlmostEqual(round3(0.625, 2), 0.62)
        self.assertAlmostEqual(round3(0.875, 2), 0.88)
        self.assertAlmostEqual(round3(-0.125, 2), -0.12)
        self.assertAlmostEqual(round3(-0.375, 2), -0.38)
        self.assertAlmostEqual(round3(-0.625, 2), -0.62)
        self.assertAlmostEqual(round3(-0.875, 2), -0.88)

        self.assertAlmostEqual(round3(0.25, 1), 0.2)
        self.assertAlmostEqual(round3(0.75, 1), 0.8)
        self.assertAlmostEqual(round3(-0.25, 1), -0.2)
        self.assertAlmostEqual(round3(-0.75, 1), -0.8)

        self.assertEqual(round3(-6.5, 0), -6.0)
        self.assertEqual(round3(-5.5, 0), -6.0)
        self.assertEqual(round3(-1.5, 0), -2.0)
        self.assertEqual(round3(-0.5, 0), 0.0)
        self.assertEqual(round3(0.5, 0), 0.0)
        self.assertEqual(round3(1.5, 0), 2.0)
        self.assertEqual(round3(2.5, 0), 2.0)
        self.assertEqual(round3(3.5, 0), 4.0)
        self.assertEqual(round3(4.5, 0), 4.0)
        self.assertEqual(round3(5.5, 0), 6.0)
        self.assertEqual(round3(6.5, 0), 6.0)

        # same but without an explicit second argument; in 2.x these
        # will give floats
        self.assertEqual(round3(-6.5), -6)
        self.assertEqual(round3(-5.5), -6)
        self.assertEqual(round3(-1.5), -2.0)
        self.assertEqual(round3(-0.5), 0)
        self.assertEqual(round3(0.5), 0)
        self.assertEqual(round3(1.5), 2)
        self.assertEqual(round3(2.5), 2)
        self.assertEqual(round3(3.5), 4)
        self.assertEqual(round3(4.5), 4)
        self.assertEqual(round3(5.5), 6)
        self.assertEqual(round3(6.5), 6)

        # no ndigits and input is already an integer: output == input
        rv = round3(1)
        self.assertEqual(rv, 1)
        self.assertTrue(isinstance(rv, int))
        rv = round3(1.0)
        self.assertEqual(rv, 1)
        self.assertTrue(isinstance(rv, int))

        self.assertEqual(round3(-25.0, -1), -20.0)
        self.assertEqual(round3(-15.0, -1), -20.0)
        self.assertEqual(round3(-5.0, -1), 0.0)
        self.assertEqual(round3(5.0, -1), 0.0)
        self.assertEqual(round3(15.0, -1), 20.0)
        self.assertEqual(round3(25.0, -1), 20.0)
        self.assertEqual(round3(35.0, -1), 40.0)
        self.assertEqual(round3(45.0, -1), 40.0)
        self.assertEqual(round3(55.0, -1), 60.0)
        self.assertEqual(round3(65.0, -1), 60.0)
        self.assertEqual(round3(75.0, -1), 80.0)
        self.assertEqual(round3(85.0, -1), 80.0)
        self.assertEqual(round3(95.0, -1), 100.0)
        self.assertEqual(round3(12325.0, -1), 12320.0)
        self.assertEqual(round3(0, -1), 0.0)

        self.assertEqual(round3(350.0, -2), 400.0)
        self.assertEqual(round3(450.0, -2), 400.0)

        self.assertAlmostEqual(round3(0.5e21, -21), 0.0)
        self.assertAlmostEqual(round3(1.5e21, -21), 2e21)
        self.assertAlmostEqual(round3(2.5e21, -21), 2e21)
        self.assertAlmostEqual(round3(5.5e21, -21), 6e21)
        self.assertAlmostEqual(round3(8.5e21, -21), 8e21)

        self.assertAlmostEqual(round3(-1.5e22, -22), -2e22)
        self.assertAlmostEqual(round3(-0.5e22, -22), 0.0)
        self.assertAlmostEqual(round3(0.5e22, -22), 0.0)
        self.assertAlmostEqual(round3(1.5e22, -22), 2e22)


NAN = float("nan")
INF = float("inf")
NINF = float("-inf")


class IsCloseTests(unittest.TestCase):
    """
    Tests taken from Python 3.5 test_math.py:
    https://hg.python.org/cpython/file/v3.5.2/Lib/test/test_math.py
    """

    isclose = staticmethod(isclose)

    def assertIsClose(self, a, b, *args, **kwargs):
        self.assertTrue(
            self.isclose(a, b, *args, **kwargs),
            msg="%s and %s should be close!" % (a, b),
        )

    def assertIsNotClose(self, a, b, *args, **kwargs):
        self.assertFalse(
            self.isclose(a, b, *args, **kwargs),
            msg="%s and %s should not be close!" % (a, b),
        )

    def assertAllClose(self, examples, *args, **kwargs):
        for a, b in examples:
            self.assertIsClose(a, b, *args, **kwargs)

    def assertAllNotClose(self, examples, *args, **kwargs):
        for a, b in examples:
            self.assertIsNotClose(a, b, *args, **kwargs)

    def test_negative_tolerances(self):
        # ValueError should be raised if either tolerance is less than zero
        with self.assertRaises(ValueError):
            self.assertIsClose(1, 1, rel_tol=-1e-100)
        with self.assertRaises(ValueError):
            self.assertIsClose(1, 1, rel_tol=1e-100, abs_tol=-1e10)

    def test_identical(self):
        # identical values must test as close
        identical_examples = [
            (2.0, 2.0),
            (0.1e200, 0.1e200),
            (1.123e-300, 1.123e-300),
            (12345, 12345.0),
            (0.0, -0.0),
            (345678, 345678),
        ]
        self.assertAllClose(identical_examples, rel_tol=0.0, abs_tol=0.0)

    def test_eight_decimal_places(self):
        # examples that are close to 1e-8, but not 1e-9
        eight_decimal_places_examples = [
            (1e8, 1e8 + 1),
            (-1e-8, -1.000000009e-8),
            (1.12345678, 1.12345679),
        ]
        self.assertAllClose(eight_decimal_places_examples, rel_tol=1e-8)
        self.assertAllNotClose(eight_decimal_places_examples, rel_tol=1e-9)

    def test_near_zero(self):
        # values close to zero
        near_zero_examples = [(1e-9, 0.0), (-1e-9, 0.0), (-1e-150, 0.0)]
        # these should not be close to any rel_tol
        self.assertAllNotClose(near_zero_examples, rel_tol=0.9)
        # these should be close to abs_tol=1e-8
        self.assertAllClose(near_zero_examples, abs_tol=1e-8)

    def test_identical_infinite(self):
        # these are close regardless of tolerance -- i.e. they are equal
        self.assertIsClose(INF, INF)
        self.assertIsClose(INF, INF, abs_tol=0.0)
        self.assertIsClose(NINF, NINF)
        self.assertIsClose(NINF, NINF, abs_tol=0.0)

    def test_inf_ninf_nan(self):
        # these should never be close (following IEEE 754 rules for equality)
        not_close_examples = [
            (NAN, NAN),
            (NAN, 1e-100),
            (1e-100, NAN),
            (INF, NAN),
            (NAN, INF),
            (INF, NINF),
            (INF, 1.0),
            (1.0, INF),
            (INF, 1e308),
            (1e308, INF),
        ]
        # use largest reasonable tolerance
        self.assertAllNotClose(not_close_examples, abs_tol=0.999999999999999)

    def test_zero_tolerance(self):
        # test with zero tolerance
        zero_tolerance_close_examples = [(1.0, 1.0), (-3.4, -3.4), (-1e-300, -1e-300)]
        self.assertAllClose(zero_tolerance_close_examples, rel_tol=0.0)

        zero_tolerance_not_close_examples = [
            (1.0, 1.000000000000001),
            (0.99999999999999, 1.0),
            (1.0e200, 0.999999999999999e200),
        ]
        self.assertAllNotClose(zero_tolerance_not_close_examples, rel_tol=0.0)

    def test_assymetry(self):
        # test the assymetry example from PEP 485
        self.assertAllClose([(9, 10), (10, 9)], rel_tol=0.1)

    def test_integers(self):
        # test with integer values
        integer_examples = [(100000001, 100000000), (123456789, 123456788)]

        self.assertAllClose(integer_examples, rel_tol=1e-8)
        self.assertAllNotClose(integer_examples, rel_tol=1e-9)

    def test_decimals(self):
        # test with Decimal values
        from decimal import Decimal

        decimal_examples = [
            (Decimal("1.00000001"), Decimal("1.0")),
            (Decimal("1.00000001e-20"), Decimal("1.0e-20")),
            (Decimal("1.00000001e-100"), Decimal("1.0e-100")),
        ]
        self.assertAllClose(decimal_examples, rel_tol=1e-8)
        self.assertAllNotClose(decimal_examples, rel_tol=1e-9)

    def test_fractions(self):
        # test with Fraction values
        from fractions import Fraction

        # could use some more examples here!
        fraction_examples = [(Fraction(1, 100000000) + 1, Fraction(1))]
        self.assertAllClose(fraction_examples, rel_tol=1e-8)
        self.assertAllNotClose(fraction_examples, rel_tol=1e-9)


class _TestRedirectStream:
    redirect_stream = None
    orig_stream = None

    def test_no_redirect_in_init(self):
        orig_stdout = getattr(sys, self.orig_stream)
        self.redirect_stream(None)
        self.assertIs(getattr(sys, self.orig_stream), orig_stdout)

    def test_redirect_to_string_io(self):
        f = StringIO()
        msg = "Consider an API like help(), which prints directly to stdout"
        orig_stdout = getattr(sys, self.orig_stream)
        with self.redirect_stream(f):
            print(msg, file=getattr(sys, self.orig_stream))
        self.assertIs(getattr(sys, self.orig_stream), orig_stdout)
        s = f.getvalue().strip()
        self.assertEqual(s, msg)

    def test_enter_result_is_target(self):
        f = StringIO()
        with self.redirect_stream(f) as enter_result:
            self.assertIs(enter_result, f)

    def test_cm_is_reusable(self):
        f = StringIO()
        write_to_f = self.redirect_stream(f)
        orig_stdout = getattr(sys, self.orig_stream)
        with write_to_f:
            print("Hello", end=" ", file=getattr(sys, self.orig_stream))
        with write_to_f:
            print("World!", file=getattr(sys, self.orig_stream))
        self.assertIs(getattr(sys, self.orig_stream), orig_stdout)
        s = f.getvalue()
        self.assertEqual(s, "Hello World!\n")

    def test_cm_is_reentrant(self):
        f = StringIO()
        write_to_f = self.redirect_stream(f)
        orig_stdout = getattr(sys, self.orig_stream)
        with write_to_f:
            print("Hello", end=" ", file=getattr(sys, self.orig_stream))
            with write_to_f:
                print("World!", file=getattr(sys, self.orig_stream))
        self.assertIs(getattr(sys, self.orig_stream), orig_stdout)
        s = f.getvalue()
        self.assertEqual(s, "Hello World!\n")


class TestRedirectStdout(_TestRedirectStream, unittest.TestCase):
    redirect_stream = redirect_stdout
    orig_stream = "stdout"


class TestRedirectStderr(_TestRedirectStream, unittest.TestCase):
    redirect_stream = redirect_stderr
    orig_stream = "stderr"


if __name__ == "__main__":
    sys.exit(unittest.main())



from fontTools.misc.testTools import parseXML
from fontTools.misc.xmlWriter import XMLWriter
from io import BytesIO
import struct
import unittest
from fontTools.ttLib import newTable


class Test_l_t_a_g(unittest.TestCase):
    DATA_ = (
        struct.pack(b">LLLHHHHHH", 1, 0, 3, 24 + 0, 2, 24 + 2, 7, 24 + 2, 2)
        + b"enzh-Hant"
    )
    TAGS_ = ["en", "zh-Hant", "zh"]

    def test_addTag(self):
        table = newTable("ltag")
        self.assertEqual(table.addTag("de-CH"), 0)
        self.assertEqual(table.addTag("gsw-LI"), 1)
        self.assertEqual(table.addTag("de-CH"), 0)
        self.assertEqual(table.tags, ["de-CH", "gsw-LI"])

    def test_decompile_compile(self):
        table = newTable("ltag")
        table.decompile(self.DATA_, ttFont=None)
        self.assertEqual(1, table.version)
        self.assertEqual(0, table.flags)
        self.assertEqual(self.TAGS_, table.tags)
        compiled = table.compile(ttFont=None)
        self.assertEqual(self.DATA_, compiled)
        self.assertIsInstance(compiled, bytes)

    def test_fromXML(self):
        table = newTable("ltag")
        for name, attrs, content in parseXML(
            '<version value="1"/>'
            '<flags value="777"/>'
            '<LanguageTag tag="sr-Latn"/>'
            '<LanguageTag tag="fa"/>'
        ):
            table.fromXML(name, attrs, content, ttFont=None)
        self.assertEqual(1, table.version)
        self.assertEqual(777, table.flags)
        self.assertEqual(["sr-Latn", "fa"], table.tags)

    def test_toXML(self):
        writer = XMLWriter(BytesIO())
        table = newTable("ltag")
        table.decompile(self.DATA_, ttFont=None)
        table.toXML(writer, ttFont=None)
        expected = (
            "\n".join(
                [
                    '<?xml version="1.0" encoding="UTF-8"?>',
                    '<version value="1"/>',
                    '<flags value="0"/>',
                    '<LanguageTag tag="en"/>',
                    '<LanguageTag tag="zh-Hant"/>',
                    '<LanguageTag tag="zh"/>',
                ]
            )
            + "\n"
        )
        self.assertEqual(expected.encode("utf_8"), writer.file.getvalue())


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



"""cff2Lib_test.py -- unit test for Adobe CFF fonts."""

from fontTools.ttLib import TTFont
from io import StringIO
import re
import os
import unittest


CURR_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
DATA_DIR = os.path.join(CURR_DIR, "data")

CFF_TTX = os.path.join(DATA_DIR, "C_F_F__2.ttx")
CFF_BIN = os.path.join(DATA_DIR, "C_F_F__2.bin")


def strip_VariableItems(string):
    # ttlib changes with the fontTools version
    string = re.sub(' ttLibVersion=".*"', "", string)
    # head table checksum and mod date changes with each save.
    string = re.sub('<checkSumAdjustment value="[^"]+"/>', "", string)
    string = re.sub('<modified value="[^"]+"/>', "", string)
    return string


class CFFTableTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        with open(CFF_BIN, "rb") as f:
            font = TTFont(file=CFF_BIN)
            cffTable = font["CFF2"]
            cls.cff2Data = cffTable.compile(font)
        with open(CFF_TTX, "r") as f:
            cff2XML = f.read()
            cff2XML = strip_VariableItems(cff2XML)
            cls.cff2XML = cff2XML.splitlines()

    def test_toXML(self):
        font = TTFont(file=CFF_BIN)
        cffTable = font["CFF2"]
        cffData = cffTable.compile(font)
        out = StringIO()
        font.saveXML(out)
        cff2XML = out.getvalue()
        cff2XML = strip_VariableItems(cff2XML)
        cff2XML = cff2XML.splitlines()
        self.assertEqual(cff2XML, self.cff2XML)

    def test_fromXML(self):
        font = TTFont(sfntVersion="OTTO")
        font.importXML(CFF_TTX)
        cffTable = font["CFF2"]
        cff2Data = cffTable.compile(font)
        self.assertEqual(cff2Data, self.cff2Data)


if __name__ == "__main__":
    unittest.main()



import os
import pytest
from fontTools.ttLib import TTFont
from fontTools.otlLib.maxContextCalc import maxCtxFont
from fontTools.feaLib.builder import addOpenTypeFeaturesFromString


def test_max_ctx_calc_no_features():
    font = TTFont()
    assert maxCtxFont(font) == 0
    font.setGlyphOrder([".notdef"])
    addOpenTypeFeaturesFromString(font, "")
    assert maxCtxFont(font) == 0


def test_max_ctx_calc_features():
    glyphs = ".notdef space A B C a b c".split()
    features = """
    lookup GSUB_EXT useExtension {
        sub a by b;
    } GSUB_EXT;

    lookup GPOS_EXT useExtension {
        pos a b -10;
    } GPOS_EXT;

    feature sub1 {
        sub A by a;
        sub A B by b;
        sub A B C by c;
        sub [A B] C by c;
        sub [A B] C [A B] by c;
        sub A' C by A B;
        sub a' by b;
        sub a' b by c;
        sub a from [A B C];
        rsub a by b;
        rsub a' by b;
        rsub a b' by c;
        rsub a b' c by A;
        rsub [a b] [a b c]' [a b] by B;
        rsub [a b] c' by A;
        lookup GSUB_EXT;
    } sub1;

    feature pos1 {
        pos A 20;
        pos A B -50;
        pos A B' 10 C;
        lookup GPOS_EXT;
    } pos1;
    """
    font = TTFont()
    font.setGlyphOrder(glyphs)
    addOpenTypeFeaturesFromString(font, features)

    assert maxCtxFont(font) == 3


@pytest.mark.parametrize(
    "file_name, max_context",
    [
        ("gsub_51", 2),
        ("gsub_52", 2),
        ("gsub_71", 1),
        ("gpos_91", 1),
    ],
)
def test_max_ctx_calc_features_ttx(file_name, max_context):
    ttx_path = os.path.join(
        os.path.dirname(__file__), "data", "{}.ttx".format(file_name)
    )
    font = TTFont()
    font.importXML(ttx_path)

    assert maxCtxFont(font) == max_context



from fontTools import ttLib
from fontTools.misc.testTools import getXML, parseXML
from fontTools.ttLib import TTFont
from fontTools.ttLib.tables.C_O_L_R_ import table_C_O_L_R_

from pathlib import Path
import binascii
import pytest


TEST_DATA_DIR = Path(__file__).parent / "data"


COLR_V0_SAMPLE = (
    (b"\x00\x00", "Version (0)"),
    (b"\x00\x01", "BaseGlyphRecordCount (1)"),
    (
        b"\x00\x00\x00\x0e",
        "Offset to BaseGlyphRecordArray from beginning of table (14)",
    ),
    (b"\x00\x00\x00\x14", "Offset to LayerRecordArray from beginning of table (20)"),
    (b"\x00\x03", "LayerRecordCount (3)"),
    (b"\x00\x06", "BaseGlyphRecord[0].BaseGlyph (6)"),
    (b"\x00\x00", "BaseGlyphRecord[0].FirstLayerIndex (0)"),
    (b"\x00\x03", "BaseGlyphRecord[0].NumLayers (3)"),
    (b"\x00\x07", "LayerRecord[0].LayerGlyph (7)"),
    (b"\x00\x00", "LayerRecord[0].PaletteIndex (0)"),
    (b"\x00\x08", "LayerRecord[1].LayerGlyph (8)"),
    (b"\x00\x01", "LayerRecord[1].PaletteIndex (1)"),
    (b"\x00\t", "LayerRecord[2].LayerGlyph (9)"),
    (b"\x00\x02", "LayerRecord[3].PaletteIndex (2)"),
)

COLR_V0_DATA = b"".join(t[0] for t in COLR_V0_SAMPLE)


COLR_V0_XML = [
    '<version value="0"/>',
    '<ColorGlyph name="glyph00006">',
    '  <layer colorID="0" name="glyph00007"/>',
    '  <layer colorID="1" name="glyph00008"/>',
    '  <layer colorID="2" name="glyph00009"/>',
    "</ColorGlyph>",
]


def dump(table, ttFont=None):
    print("\n".join(getXML(table.toXML, ttFont)))


def diff_binary_fragments(font_bytes, expected_fragments):
    pos = 0
    prev_desc = ""
    errors = 0
    for expected_bytes, description in expected_fragments:
        actual_bytes = font_bytes[pos : pos + len(expected_bytes)]
        if actual_bytes != expected_bytes:
            print(
                f'{description} (previous "{prev_desc}", actual_bytes: {"".join("%02x" % v for v in actual_bytes)} bytes: {str(font_bytes[pos:pos+16])}'
            )
            errors += 1
        pos += len(expected_bytes)
        prev_desc = description
    assert errors == 0
    assert pos == len(
        font_bytes
    ), f"Leftover font bytes, used {pos} of {len(font_bytes)}"


@pytest.fixture
def font():
    font = ttLib.TTFont()
    font.setGlyphOrder(["glyph%05d" % i for i in range(30)])
    return font


class COLR_V0_Test(object):
    def test_decompile_and_compile(self, font):
        colr = table_C_O_L_R_()
        colr.decompile(COLR_V0_DATA, font)
        diff_binary_fragments(colr.compile(font), COLR_V0_SAMPLE)

    def test_decompile_and_dump_xml(self, font):
        colr = table_C_O_L_R_()
        colr.decompile(COLR_V0_DATA, font)

        dump(colr, font)
        assert getXML(colr.toXML, font) == COLR_V0_XML

    def test_load_from_xml_and_compile(self, font):
        colr = table_C_O_L_R_()
        for name, attrs, content in parseXML(COLR_V0_XML):
            colr.fromXML(name, attrs, content, font)

        diff_binary_fragments(colr.compile(font), COLR_V0_SAMPLE)

    def test_round_trip_xml(self, font):
        colr = table_C_O_L_R_()
        for name, attrs, content in parseXML(COLR_V0_XML):
            colr.fromXML(name, attrs, content, font)
        compiled = colr.compile(font)

        colr = table_C_O_L_R_()
        colr.decompile(compiled, font)
        assert getXML(colr.toXML, font) == COLR_V0_XML


COLR_V1_SAMPLE = (
    (b"\x00\x01", "Version (1)"),
    (b"\x00\x01", "BaseGlyphRecordCount (1)"),
    (
        b"\x00\x00\x00\x22",
        "Offset to BaseGlyphRecordArray from beginning of table (34)",
    ),
    (b"\x00\x00\x00\x28", "Offset to LayerRecordArray from beginning of table (40)"),
    (b"\x00\x03", "LayerRecordCount (3)"),
    (b"\x00\x00\x00\x34", "Offset to BaseGlyphList from beginning of table (52)"),
    (b"\x00\x00\x00\x9f", "Offset to LayerList from beginning of table (159)"),
    (b"\x00\x00\x01\x66", "Offset to ClipList (358)"),
    (b"\x00\x00\x00\x00", "Offset to DeltaSetIndexMap (NULL)"),
    (b"\x00\x00\x00\x00", "Offset to VarStore (NULL)"),
    (b"\x00\x06", "BaseGlyphRecord[0].BaseGlyph (6)"),
    (b"\x00\x00", "BaseGlyphRecord[0].FirstLayerIndex (0)"),
    (b"\x00\x03", "BaseGlyphRecord[0].NumLayers (3)"),
    (b"\x00\x07", "LayerRecord[0].LayerGlyph (7)"),
    (b"\x00\x00", "LayerRecord[0].PaletteIndex (0)"),
    (b"\x00\x08", "LayerRecord[1].LayerGlyph (8)"),
    (b"\x00\x01", "LayerRecord[1].PaletteIndex (1)"),
    (b"\x00\t", "LayerRecord[2].LayerGlyph (9)"),
    (b"\x00\x02", "LayerRecord[2].PaletteIndex (2)"),
    # BaseGlyphList
    (b"\x00\x00\x00\x03", "BaseGlyphList.BaseGlyphCount (3)"),
    (b"\x00\n", "BaseGlyphList.BaseGlyphPaintRecord[0].BaseGlyph (10)"),
    (
        b"\x00\x00\x00\x16",
        "Offset to Paint table from beginning of BaseGlyphList (22)",
    ),
    (b"\x00\x0e", "BaseGlyphList.BaseGlyphPaintRecord[1].BaseGlyph (14)"),
    (
        b"\x00\x00\x00\x1c",
        "Offset to Paint table from beginning of BaseGlyphList (28)",
    ),
    (b"\x00\x0f", "BaseGlyphList.BaseGlyphPaintRecord[2].BaseGlyph (15)"),
    (
        b"\x00\x00\x00\x4a",
        "Offset to Paint table from beginning of BaseGlyphList (74)",
    ),
    # BaseGlyphPaintRecord[0]
    (b"\x01", "BaseGlyphPaintRecord[0].Paint.Format (1)"),
    (b"\x04", "BaseGlyphPaintRecord[0].Paint.NumLayers (4)"),
    (b"\x00\x00\x00\x00", "BaseGlyphPaintRecord[0].Paint.FirstLayerIndex (0)"),
    # BaseGlyphPaintRecord[1]
    (b"\x20", "BaseGlyphPaintRecord[1].Paint.Format (32)"),
    (b"\x00\x00\x0f", "Offset to SourcePaint from beginning of PaintComposite (15)"),
    (b"\x03", "BaseGlyphPaintRecord[1].Paint.CompositeMode [SRC_OVER] (3)"),
    (b"\x00\x00\x08", "Offset to BackdropPaint from beginning of PaintComposite (8)"),
    (b"\x0d", "BaseGlyphPaintRecord[1].Paint.BackdropPaint.Format (13)"),
    (b"\x00\x00\x07", "Offset to Paint from beginning of PaintVarTransform (7)"),
    (
        b"\x00\x00\x0a",
        "Offset to VarAffine2x3 from beginning of PaintVarTransform (10)",
    ),
    (b"\x0b", "BaseGlyphPaintRecord[1].Paint.BackdropPaint.Format (11)"),
    (b"\x00\x0a", "BaseGlyphPaintRecord[1].Paint.BackdropPaint.Glyph (10)"),
    (b"\x00\x01\x00\x00", "VarAffine2x3.xx (1.0)"),
    (b"\x00\x00\x00\x00", "VarAffine2x3.xy (0.0)"),
    (b"\x00\x00\x00\x00", "VarAffine2x3.yx (0.0)"),
    (b"\x00\x01\x00\x00", "VarAffine2x3.yy (1.0)"),
    (b"\x01\x2c\x00\x00", "VarAffine2x3.dx (300.0)"),
    (b"\x00\x00\x00\x00", "VarAffine2x3.dy (0.0)"),
    (b"\x00\x00\x00\x00", "VarIndexBase (0)"),
    (b"\x0a", "BaseGlyphPaintRecord[1].Paint.SourcePaint.Format (10)"),
    (b"\x00\x00\x06", "Offset to Paint subtable from beginning of PaintGlyph (6)"),
    (b"\x00\x0b", "BaseGlyphPaintRecord[1].Paint.SourcePaint.Glyph (11)"),
    (b"\x08", "BaseGlyphPaintRecord[1].Paint.SourcePaint.Paint.Format (8)"),
    (b"\x00\x00\x0c", "Offset to ColorLine from beginning of PaintSweepGradient (12)"),
    (b"\x01\x03", "centerX (259)"),
    (b"\x01\x2c", "centerY (300)"),
    (b"\x10\x00", "startAngle (0.25)"),
    (b"\x30\x00", "endAngle (0.75)"),
    (b"\x00", "ColorLine.Extend (0; pad)"),
    (b"\x00\x02", "ColorLine.StopCount (2)"),
    (b"\x00\x00", "ColorLine.ColorStop[0].StopOffset (0.0)"),
    (b"\x00\x03", "ColorLine.ColorStop[0].PaletteIndex (3)"),
    (b"@\x00", "ColorLine.ColorStop[0].Alpha (1.0)"),
    (b"@\x00", "ColorLine.ColorStop[1].StopOffset (1.0)"),
    (b"\x00\x05", "ColorLine.ColorStop[1].PaletteIndex (5)"),
    (b"@\x00", "ColorLine.ColorStop[1].Alpha (1.0)"),
    # LayerList
    (b"\x00\x00\x00\x05", "LayerList.LayerCount (5)"),
    (
        b"\x00\x00\x00\x18",
        "First Offset to Paint table from beginning of LayerList (24)",
    ),
    (
        b"\x00\x00\x00\x27",
        "Second Offset to Paint table from beginning of LayerList (39)",
    ),
    (
        b"\x00\x00\x00\x52",
        "Third Offset to Paint table from beginning of LayerList (82)",
    ),
    (
        b"\x00\x00\x00\xa2",
        "Fourth Offset to Paint table from beginning of LayerList (162)",
    ),
    (
        b"\x00\x00\x00\xbc",
        "Fifth Offset to Paint table from beginning of LayerList (188)",
    ),
    # BaseGlyphPaintRecord[2]
    (b"\x0a", "BaseGlyphPaintRecord[2].Paint.Format (10)"),
    (b"\x00\x00\x06", "Offset to Paint subtable from beginning of PaintGlyph (6)"),
    (b"\x00\x0b", "BaseGlyphPaintRecord[2].Paint.Glyph (11)"),
    # PaintVarSolid
    (b"\x03", "LayerList.Paint[0].Paint.Format (3)"),
    (b"\x00\x02", "Paint.PaletteIndex (2)"),
    (b" \x00", "Paint.Alpha.value (0.5)"),
    (b"\x00\x00\x00\x06", "VarIndexBase (6)"),
    # PaintGlyph glyph00012
    (b"\x0a", "LayerList.Paint[1].Format (10)"),
    (b"\x00\x00\x06", "Offset to Paint subtable from beginning of PaintGlyph (6)"),
    (b"\x00\x0c", "LayerList.Paint[1].Glyph (glyph00012)"),
    (b"\x04", "LayerList.Paint[1].Paint.Format (4)"),
    (b"\x00\x00\x10", "Offset to ColorLine from beginning of PaintLinearGradient (16)"),
    (b"\x00\x01", "Paint.x0 (1)"),
    (b"\x00\x02", "Paint.y0 (2)"),
    (b"\xff\xfd", "Paint.x1 (-3)"),
    (b"\xff\xfc", "Paint.y1 (-4)"),
    (b"\x00\x05", "Paint.x2 (5)"),
    (b"\x00\x06", "Paint.y2 (6)"),
    (b"\x01", "ColorLine.Extend (1; repeat)"),
    (b"\x00\x03", "ColorLine.StopCount (3)"),
    (b"\x00\x00", "ColorLine.ColorStop[0].StopOffset (0.0)"),
    (b"\x00\x03", "ColorLine.ColorStop[0].PaletteIndex (3)"),
    (b"@\x00", "ColorLine.ColorStop[0].Alpha (1.0)"),
    (b" \x00", "ColorLine.ColorStop[1].StopOffset (0.5)"),
    (b"\x00\x04", "ColorLine.ColorStop[1].PaletteIndex (4)"),
    (b"@\x00", "ColorLine.ColorStop[1].Alpha (1.0)"),
    (b"@\x00", "ColorLine.ColorStop[2].StopOffset (1.0)"),
    (b"\x00\x05", "ColorLine.ColorStop[2].PaletteIndex (5)"),
    (b"@\x00", "ColorLine.ColorStop[2].Alpha (1.0)"),
    # PaintGlyph glyph00013
    (b"\x0a", "LayerList.Paint[2].Format (10)"),
    (b"\x00\x00\x06", "Offset to Paint subtable from beginning of PaintGlyph (6)"),
    (b"\x00\x0d", "LayerList.Paint[2].Glyph (13)"),
    (b"\x0c", "LayerList.Paint[2].Paint.Format (12)"),
    (b"\x00\x00\x07", "Offset to Paint subtable from beginning of PaintTransform (7)"),
    (
        b"\x00\x00\x32",
        "Offset to Affine2x3 subtable from beginning of PaintTransform (50)",
    ),
    (b"\x07", "LayerList.Paint[2].Paint.Paint.Format (7)"),
    (
        b"\x00\x00\x14",
        "Offset to ColorLine from beginning of PaintVarRadialGradient (20)",
    ),
    (b"\x00\x07", "Paint.x0.value (7)"),
    (b"\x00\x08", "Paint.y0.value (8)"),
    (b"\x00\t", "Paint.r0.value (9)"),
    (b"\x00\n", "Paint.x1.value (10)"),
    (b"\x00\x0b", "Paint.y1.value (11)"),
    (b"\x00\x0c", "Paint.r1.value (12)"),
    (b"\xff\xff\xff\xff", "VarIndexBase (0xFFFFFFFF)"),
    (b"\x00", "ColorLine.Extend (0; pad)"),
    (b"\x00\x02", "ColorLine.StopCount (2)"),
    (b"\x00\x00", "ColorLine.ColorStop[0].StopOffset.value (0.0)"),
    (b"\x00\x06", "ColorLine.ColorStop[0].PaletteIndex (6)"),
    (b"@\x00", "ColorLine.ColorStop[0].Alpha.value (1.0)"),
    (b"\xff\xff\xff\xff", "VarIndexBase (0xFFFFFFFF)"),
    (b"@\x00", "ColorLine.ColorStop[1].StopOffset.value (1.0)"),
    (b"\x00\x07", "ColorLine.ColorStop[1].PaletteIndex (7)"),
    (b"\x19\x9a", "ColorLine.ColorStop[1].Alpha.value (0.4)"),
    (b"\x00\x00\x00\x07", "VarIndexBase (7)"),
    (b"\xff\xf3\x00\x00", "Affine2x3.xx (-13)"),
    (b"\x00\x0e\x00\x00", "Affine2x3.xy (14)"),
    (b"\x00\x0f\x00\x00", "Affine2x3.yx (15)"),
    (b"\xff\xef\x00\x00", "Affine2x3.yy (-17)"),
    (b"\x00\x12\x00\x00", "Affine2x3.yy (18)"),
    (b"\x00\x13\x00\x00", "Affine2x3.yy (19)"),
    # PaintTranslate
    (b"\x0e", "LayerList.Paint[3].Format (14)"),
    (b"\x00\x00\x08", "Offset to Paint subtable from beginning of PaintTranslate (8)"),
    (b"\x01\x01", "dx (257)"),
    (b"\x01\x02", "dy (258)"),
    # PaintRotateAroundCenter
    (b"\x1a", "LayerList.Paint[3].Paint.Format (26)"),
    (
        b"\x00\x00\x0a",
        "Offset to Paint subtable from beginning of PaintRotateAroundCenter (11)",
    ),
    (b"\x10\x00", "angle (0.25)"),
    (b"\x00\xff", "centerX (255)"),
    (b"\x01\x00", "centerY (256)"),
    # PaintSkew
    (b"\x1c", "LayerList.Paint[3].Paint.Paint.Format (28)"),
    (
        b"\x00\x00\x08",
        "Offset to Paint subtable from beginning of PaintSkew (8)",
    ),
    (b"\xfc\x17", "xSkewAngle (-0.0611)"),
    (b"\x01\xc7", "ySkewAngle (0.0278)"),
    # PaintGlyph glyph00011 (pointed to by both PaintSkew above and by LayerList[4] offset)
    (b"\x0a", "LayerList.Paint[3].Paint.Paint.Paint.Format (10)"),
    (b"\x00\x00\x06", "Offset to Paint subtable from beginning of PaintGlyph (6)"),
    (b"\x00\x0b", "LayerList.Paint[2].Glyph (11)"),
    # PaintSolid
    (b"\x02", "LayerList.Paint[0].Paint.Paint.Paint.Paint.Format (2)"),
    (b"\x00\x02", "Paint.PaletteIndex (2)"),
    (b" \x00", "Paint.Alpha (0.5)"),
    # ClipList
    (b"\x01", "ClipList.Format (1)"),
    (b"\x00\x00\x00\x02", "ClipList.ClipCount (2)"),
    (b"\x00\x0a", "ClipRecord[0].StartGlyphID (10)"),
    (b"\x00\x0a", "ClipRecord[0].EndGlyphID (10)"),
    (b"\x00\x00\x13", "Offset to ClipBox subtable from beginning of ClipList (19)"),
    (b"\x00\x0e", "ClipRecord[1].StartGlyphID (14)"),
    (b"\x00\x0f", "ClipRecord[1].EndGlyphID (15)"),
    (b"\x00\x00\x20", "Offset to ClipBox subtable from beginning of ClipList (32)"),
    (b"\x02", "ClipBox.Format (2)"),
    (b"\x00\x00", "ClipBox.xMin (0)"),
    (b"\x00\x00", "ClipBox.yMin (0)"),
    (b"\x01\xf4", "ClipBox.xMax (500)"),
    (b"\x01\xf4", "ClipBox.yMax (500)"),
    (b"\x00\x00\x00\t", "ClipBox.VarIndexBase (9)"),
    (b"\x01", "ClipBox.Format (1)"),
    (b"\x00\x00", "ClipBox.xMin (0)"),
    (b"\x00\x00", "ClipBox.yMin (0)"),
    (b"\x03\xe8", "ClipBox.xMax (1000)"),
    (b"\x03\xe8", "ClipBox.yMax (1000)"),
)

COLR_V1_DATA = b"".join(t[0] for t in COLR_V1_SAMPLE)

COLR_V1_XML = [
    '<Version value="1"/>',
    "<!-- BaseGlyphRecordCount=1 -->",
    "<BaseGlyphRecordArray>",
    '  <BaseGlyphRecord index="0">',
    '    <BaseGlyph value="glyph00006"/>',
    '    <FirstLayerIndex value="0"/>',
    '    <NumLayers value="3"/>',
    "  </BaseGlyphRecord>",
    "</BaseGlyphRecordArray>",
    "<LayerRecordArray>",
    '  <LayerRecord index="0">',
    '    <LayerGlyph value="glyph00007"/>',
    '    <PaletteIndex value="0"/>',
    "  </LayerRecord>",
    '  <LayerRecord index="1">',
    '    <LayerGlyph value="glyph00008"/>',
    '    <PaletteIndex value="1"/>',
    "  </LayerRecord>",
    '  <LayerRecord index="2">',
    '    <LayerGlyph value="glyph00009"/>',
    '    <PaletteIndex value="2"/>',
    "  </LayerRecord>",
    "</LayerRecordArray>",
    "<!-- LayerRecordCount=3 -->",
    "<BaseGlyphList>",
    "  <!-- BaseGlyphCount=3 -->",
    '  <BaseGlyphPaintRecord index="0">',
    '    <BaseGlyph value="glyph00010"/>',
    '    <Paint Format="1"><!-- PaintColrLayers -->',
    '      <NumLayers value="4"/>',
    '      <FirstLayerIndex value="0"/>',
    "    </Paint>",
    "  </BaseGlyphPaintRecord>",
    '  <BaseGlyphPaintRecord index="1">',
    '    <BaseGlyph value="glyph00014"/>',
    '    <Paint Format="32"><!-- PaintComposite -->',
    '      <SourcePaint Format="11"><!-- PaintColrGlyph -->',
    '        <Glyph value="glyph00010"/>',
    "      </SourcePaint>",
    '      <CompositeMode value="src_over"/>',
    '      <BackdropPaint Format="13"><!-- PaintVarTransform -->',
    '        <Paint Format="11"><!-- PaintColrGlyph -->',
    '          <Glyph value="glyph00010"/>',
    "        </Paint>",
    "        <Transform>",
    '          <xx value="1.0"/>',
    '          <yx value="0.0"/>',
    '          <xy value="0.0"/>',
    '          <yy value="1.0"/>',
    '          <dx value="300.0"/>',
    '          <dy value="0.0"/>',
    '          <VarIndexBase value="0"/>',
    "        </Transform>",
    "      </BackdropPaint>",
    "    </Paint>",
    "  </BaseGlyphPaintRecord>",
    '  <BaseGlyphPaintRecord index="2">',
    '    <BaseGlyph value="glyph00015"/>',
    '    <Paint Format="10"><!-- PaintGlyph -->',
    '      <Paint Format="8"><!-- PaintSweepGradient -->',
    "        <ColorLine>",
    '          <Extend value="pad"/>',
    "          <!-- StopCount=2 -->",
    '          <ColorStop index="0">',
    '            <StopOffset value="0.0"/>',
    '            <PaletteIndex value="3"/>',
    '            <Alpha value="1.0"/>',
    "          </ColorStop>",
    '          <ColorStop index="1">',
    '            <StopOffset value="1.0"/>',
    '            <PaletteIndex value="5"/>',
    '            <Alpha value="1.0"/>',
    "          </ColorStop>",
    "        </ColorLine>",
    '        <centerX value="259"/>',
    '        <centerY value="300"/>',
    '        <startAngle value="225.0"/>',
    '        <endAngle value="315.0"/>',
    "      </Paint>",
    '      <Glyph value="glyph00011"/>',
    "    </Paint>",
    "  </BaseGlyphPaintRecord>",
    "</BaseGlyphList>",
    "<LayerList>",
    "  <!-- LayerCount=5 -->",
    '  <Paint index="0" Format="10"><!-- PaintGlyph -->',
    '    <Paint Format="3"><!-- PaintVarSolid -->',
    '      <PaletteIndex value="2"/>',
    '      <Alpha value="0.5"/>',
    '      <VarIndexBase value="6"/>',
    "    </Paint>",
    '    <Glyph value="glyph00011"/>',
    "  </Paint>",
    '  <Paint index="1" Format="10"><!-- PaintGlyph -->',
    '    <Paint Format="4"><!-- PaintLinearGradient -->',
    "      <ColorLine>",
    '        <Extend value="repeat"/>',
    "        <!-- StopCount=3 -->",
    '        <ColorStop index="0">',
    '          <StopOffset value="0.0"/>',
    '          <PaletteIndex value="3"/>',
    '          <Alpha value="1.0"/>',
    "        </ColorStop>",
    '        <ColorStop index="1">',
    '          <StopOffset value="0.5"/>',
    '          <PaletteIndex value="4"/>',
    '          <Alpha value="1.0"/>',
    "        </ColorStop>",
    '        <ColorStop index="2">',
    '          <StopOffset value="1.0"/>',
    '          <PaletteIndex value="5"/>',
    '          <Alpha value="1.0"/>',
    "        </ColorStop>",
    "      </ColorLine>",
    '      <x0 value="1"/>',
    '      <y0 value="2"/>',
    '      <x1 value="-3"/>',
    '      <y1 value="-4"/>',
    '      <x2 value="5"/>',
    '      <y2 value="6"/>',
    "    </Paint>",
    '    <Glyph value="glyph00012"/>',
    "  </Paint>",
    '  <Paint index="2" Format="10"><!-- PaintGlyph -->',
    '    <Paint Format="12"><!-- PaintTransform -->',
    '      <Paint Format="7"><!-- PaintVarRadialGradient -->',
    "        <ColorLine>",
    '          <Extend value="pad"/>',
    "          <!-- StopCount=2 -->",
    '          <ColorStop index="0">',
    '            <StopOffset value="0.0"/>',
    '            <PaletteIndex value="6"/>',
    '            <Alpha value="1.0"/>',
    "            <VarIndexBase/>",
    "          </ColorStop>",
    '          <ColorStop index="1">',
    '            <StopOffset value="1.0"/>',
    '            <PaletteIndex value="7"/>',
    '            <Alpha value="0.4"/>',
    '            <VarIndexBase value="7"/>',
    "          </ColorStop>",
    "        </ColorLine>",
    '        <x0 value="7"/>',
    '        <y0 value="8"/>',
    '        <r0 value="9"/>',
    '        <x1 value="10"/>',
    '        <y1 value="11"/>',
    '        <r1 value="12"/>',
    "        <VarIndexBase/>",
    "      </Paint>",
    "      <Transform>",
    '        <xx value="-13.0"/>',
    '        <yx value="14.0"/>',
    '        <xy value="15.0"/>',
    '        <yy value="-17.0"/>',
    '        <dx value="18.0"/>',
    '        <dy value="19.0"/>',
    "      </Transform>",
    "    </Paint>",
    '    <Glyph value="glyph00013"/>',
    "  </Paint>",
    '  <Paint index="3" Format="14"><!-- PaintTranslate -->',
    '    <Paint Format="26"><!-- PaintRotateAroundCenter -->',
    '      <Paint Format="28"><!-- PaintSkew -->',
    '        <Paint Format="10"><!-- PaintGlyph -->',
    '          <Paint Format="2"><!-- PaintSolid -->',
    '            <PaletteIndex value="2"/>',
    '            <Alpha value="0.5"/>',
    "          </Paint>",
    '          <Glyph value="glyph00011"/>',
    "        </Paint>",
    '        <xSkewAngle value="-11.0"/>',
    '        <ySkewAngle value="5.0"/>',
    "      </Paint>",
    '      <angle value="45.0"/>',
    '      <centerX value="255"/>',
    '      <centerY value="256"/>',
    "    </Paint>",
    '    <dx value="257"/>',
    '    <dy value="258"/>',
    "  </Paint>",
    '  <Paint index="4" Format="10"><!-- PaintGlyph -->',
    '    <Paint Format="2"><!-- PaintSolid -->',
    '      <PaletteIndex value="2"/>',
    '      <Alpha value="0.5"/>',
    "    </Paint>",
    '    <Glyph value="glyph00011"/>',
    "  </Paint>",
    "</LayerList>",
    '<ClipList Format="1">',
    "  <Clip>",
    '    <Glyph value="glyph00010"/>',
    '    <ClipBox Format="2">',
    '      <xMin value="0"/>',
    '      <yMin value="0"/>',
    '      <xMax value="500"/>',
    '      <yMax value="500"/>',
    '      <VarIndexBase value="9"/>',
    "    </ClipBox>",
    "  </Clip>",
    "  <Clip>",
    '    <Glyph value="glyph00014"/>',
    '    <Glyph value="glyph00015"/>',
    '    <ClipBox Format="1">',
    '      <xMin value="0"/>',
    '      <yMin value="0"/>',
    '      <xMax value="1000"/>',
    '      <yMax value="1000"/>',
    "    </ClipBox>",
    "  </Clip>",
    "</ClipList>",
]

COLR_V1_VAR_XML = [
    '<VarIndexMap Format="0">',
    "  <!-- Omitted values default to 0xFFFF/0xFFFF (no variations) -->",
    '  <Map index="0" outer="1" inner="0"/>',
    '  <Map index="1"/>',
    '  <Map index="2"/>',
    '  <Map index="3" outer="1" inner="0"/>',
    '  <Map index="4"/>',
    '  <Map index="5"/>',
    '  <Map index="6" outer="0" inner="2"/>',
    '  <Map index="7" outer="0" inner="0"/>',
    '  <Map index="8" outer="0" inner="1"/>',
    '  <Map index="9"/>',
    '  <Map index="10"/>',
    '  <Map index="11" outer="0" inner="3"/>',
    '  <Map index="12" outer="0" inner="3"/>',
    "</VarIndexMap>",
    '<VarStore Format="1">',
    '  <Format value="1"/>',
    "  <VarRegionList>",
    "    <!-- RegionAxisCount=1 -->",
    "    <!-- RegionCount=1 -->",
    '    <Region index="0">',
    '      <VarRegionAxis index="0">',
    '        <StartCoord value="0.0"/>',
    '        <PeakCoord value="1.0"/>',
    '        <EndCoord value="1.0"/>',
    "      </VarRegionAxis>",
    "    </Region>",
    "  </VarRegionList>",
    "  <!-- VarDataCount=2 -->",
    '  <VarData index="0">',
    "    <!-- ItemCount=4 -->",
    '    <NumShorts value="1"/>',
    "    <!-- VarRegionCount=1 -->",
    '    <VarRegionIndex index="0" value="0"/>',
    '    <Item index="0" value="[-3277]"/>',
    '    <Item index="1" value="[6553]"/>',
    '    <Item index="2" value="[8192]"/>',
    '    <Item index="3" value="[500]"/>',
    "  </VarData>",
    '  <VarData index="1">',
    "    <!-- ItemCount=1 -->",
    '    <NumShorts value="32769"/>',
    "    <!-- VarRegionCount=1 -->",
    '    <VarRegionIndex index="0" value="0"/>',
    '    <Item index="0" value="[65536]"/>',
    "  </VarData>",
    "</VarStore>",
]


class COLR_V1_Test(object):
    def test_decompile_and_compile(self, font):
        colr = table_C_O_L_R_()
        colr.decompile(COLR_V1_DATA, font)
        diff_binary_fragments(colr.compile(font), COLR_V1_SAMPLE)

    def test_decompile_and_dump_xml(self, font):
        colr = table_C_O_L_R_()
        colr.decompile(COLR_V1_DATA, font)

        dump(colr, font)
        assert getXML(colr.toXML, font) == COLR_V1_XML

    def test_load_from_xml_and_compile(self, font):
        colr = table_C_O_L_R_()
        for name, attrs, content in parseXML(COLR_V1_XML):
            colr.fromXML(name, attrs, content, font)
        diff_binary_fragments(colr.compile(font), COLR_V1_SAMPLE)

    def test_round_trip_xml(self, font):
        colr = table_C_O_L_R_()
        for name, attrs, content in parseXML(COLR_V1_XML):
            colr.fromXML(name, attrs, content, font)
        compiled = colr.compile(font)

        colr = table_C_O_L_R_()
        colr.decompile(compiled, font)
        assert getXML(colr.toXML, font) == COLR_V1_XML

    @pytest.mark.parametrize("quantization", [1, 10, 100])
    @pytest.mark.parametrize("flavor", ["glyf", "cff"])
    def test_computeClipBoxes(self, flavor, quantization):
        font = TTFont()
        font.importXML(TEST_DATA_DIR / f"COLRv1-clip-boxes-{flavor}.ttx")
        assert font["COLR"].table.ClipList is None

        font["COLR"].table.computeClipBoxes(font.getGlyphSet(), quantization)

        clipList = font["COLR"].table.ClipList
        assert len(clipList.clips) > 0

        expected = TTFont()
        expected.importXML(
            TEST_DATA_DIR / f"COLRv1-clip-boxes-q{quantization}-expected.ttx"
        )
        expectedClipList = expected["COLR"].table.ClipList

        assert getXML(clipList.toXML) == getXML(expectedClipList.toXML)


class COLR_V1_Variable_Test(object):
    def test_round_trip_xml(self, font):
        colr = table_C_O_L_R_()
        xml = COLR_V1_XML + COLR_V1_VAR_XML
        for name, attrs, content in parseXML(xml):
            colr.fromXML(name, attrs, content, font)
        compiled = colr.compile(font)

        colr = table_C_O_L_R_()
        colr.decompile(compiled, font)
        assert getXML(colr.toXML, font) == xml



import gzip
import io
import struct

from fontTools.misc import etree
from fontTools.misc.testTools import getXML, parseXML
from fontTools.ttLib import TTFont
from fontTools.ttLib.tables.S_V_G_ import table_S_V_G_

import pytest


def dump(table, ttFont=None):
    print("\n".join(getXML(table.toXML, ttFont)))


def compress(data: bytes) -> bytes:
    buf = io.BytesIO()
    with gzip.GzipFile(None, "w", fileobj=buf, mtime=0) as gz:
        gz.write(data)
    return buf.getvalue()


def strip_xml_whitespace(xml_string):
    def strip_or_none(text):
        text = text.strip() if text else None
        return text if text else None

    tree = etree.fromstring(xml_string)
    for e in tree.iter("*"):
        e.text = strip_or_none(e.text)
        e.tail = strip_or_none(e.tail)
    return etree.tostring(tree, encoding="utf-8")


SVG_DOCS = [
    strip_xml_whitespace(svg)
    for svg in (
        b"""\
        <svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" version="1.1">
          <defs>
            <rect x="100" y="-200" width="300" height="400" id="p1"/>
          </defs>
          <g id="glyph1">
            <use xlink:href="#p1" fill="#red"/>
          </g>
          <g id="glyph2">
            <use xlink:href="#p1" fill="#blue"/>
          </g>
          <g id="glyph4">
            <use xlink:href="#p1" fill="#green"/>
          </g>
        </svg>""",
        b"""\
        <svg xmlns="http://www.w3.org/2000/svg" version="1.1">
          <g id="glyph3">
            <path d="M0,0 L100,0 L50,100 Z" fill="#red"/>
            <path d="M10,10 L110,10 L60,110 Z" fill="#blue"/>
            <path d="M20,20 L120,20 L70,120 Z" fill="#green"/>
          </g>
        </svg>""",
    )
]


OTSVG_DATA = b"".join(
    [
        # SVG table header
        b"\x00\x00"  # version (0)
        b"\x00\x00\x00\x0a"  # offset to SVGDocumentList (10)
        b"\x00\x00\x00\x00"  # reserved (0)
        #  SVGDocumentList
        b"\x00\x03"  # number of SVGDocumentRecords (3)
        # SVGDocumentRecord[0]
        b"\x00\x01"  # startGlyphID (1)
        b"\x00\x02"  # endGlyphID (2)
        b"\x00\x00\x00\x26"  # svgDocOffset (2 + 12*3 == 38 == 0x26)
        + struct.pack(">L", len(SVG_DOCS[0]))  # svgDocLength
        # SVGDocumentRecord[1] (compressed)
        + b"\x00\x03"  # startGlyphID (3)
        b"\x00\x03"  # endGlyphID (3)
        + struct.pack(">L", 0x26 + len(SVG_DOCS[0]))  # svgDocOffset
        + struct.pack(">L", len(compress(SVG_DOCS[1])))  # svgDocLength
        # SVGDocumentRecord[2]
        + b"\x00\x04"  # startGlyphID (4)
        b"\x00\x04"  # endGlyphID (4)
        b"\x00\x00\x00\x26"  # svgDocOffset (38); records 0 and 2 point to same SVG doc
        + struct.pack(">L", len(SVG_DOCS[0]))  # svgDocLength
    ]
    + [SVG_DOCS[0], compress(SVG_DOCS[1])]
)

OTSVG_TTX = [
    '<svgDoc endGlyphID="2" startGlyphID="1">',
    f"  <![CDATA[{SVG_DOCS[0].decode()}]]>",
    "</svgDoc>",
    '<svgDoc compressed="1" endGlyphID="3" startGlyphID="3">',
    f"  <![CDATA[{SVG_DOCS[1].decode()}]]>",
    "</svgDoc>",
    '<svgDoc endGlyphID="4" startGlyphID="4">',
    f"  <![CDATA[{SVG_DOCS[0].decode()}]]>",
    "</svgDoc>",
]


@pytest.fixture
def font():
    font = TTFont()
    font.setGlyphOrder([".notdef"] + ["glyph%05d" % i for i in range(1, 30)])
    return font


def test_decompile_and_compile(font):
    table = table_S_V_G_()
    table.decompile(OTSVG_DATA, font)
    assert table.compile(font) == OTSVG_DATA


def test_decompile_and_dump_ttx(font):
    table = table_S_V_G_()
    table.decompile(OTSVG_DATA, font)

    dump(table, font)
    assert getXML(table.toXML, font) == OTSVG_TTX


def test_load_from_ttx_and_compile(font):
    table = table_S_V_G_()
    for name, attrs, content in parseXML(OTSVG_TTX):
        table.fromXML(name, attrs, content, font)
    assert table.compile(font) == OTSVG_DATA


def test_round_trip_ttx(font):
    table = table_S_V_G_()
    for name, attrs, content in parseXML(OTSVG_TTX):
        table.fromXML(name, attrs, content, font)
    compiled = table.compile(font)

    table = table_S_V_G_()
    table.decompile(compiled, font)
    assert getXML(table.toXML, font) == OTSVG_TTX


def test_unpack_svg_doc_as_3_tuple():
    # test that the legacy docList as list of 3-tuples interface still works
    # even after the new SVGDocument class with extra `compressed` attribute
    # was added
    table = table_S_V_G_()
    table.decompile(OTSVG_DATA, font)

    for doc, compressed in zip(table.docList, (False, True, False)):
        assert len(doc) == 3
        data, startGID, endGID = doc
        assert doc.data == data
        assert doc.startGlyphID == startGID
        assert doc.endGlyphID == endGID
        assert doc.compressed == compressed



import pytest
import random
from io import StringIO
from fontTools.misc.xmlWriter import XMLWriter
from fontTools.misc.roundTools import noRound
from fontTools.varLib.models import VariationModel
from fontTools.varLib.varStore import OnlineVarStoreBuilder, VarStoreInstancer
from fontTools.ttLib import TTFont, newTable
from fontTools.ttLib.tables._f_v_a_r import Axis
from fontTools.ttLib.tables.otBase import OTTableReader, OTTableWriter
from fontTools.ttLib.tables.otTables import VarStore


@pytest.mark.parametrize(
    "locations, masterValues",
    [
        (
            [{}, {"a": 1}],
            [
                [10, 10],  # Test NO_VARIATION_INDEX
                [100, 2000],
                [100, 22000],
            ],
        ),
        (
            [{}, {"a": 1}, {"b": 1}, {"a": 1, "b": 1}],
            [
                [10, 20, 40, 60],
                [100, 2000, 400, 6000],
                [7100, 22000, 4000, 30000],
            ],
        ),
        (
            [{}, {"a": 1}],
            [
                [10, 20],
                [42000, 100],
                [100, 52000],
            ],
        ),
        (
            [{}, {"a": 1}, {"b": 1}, {"a": 1, "b": 1}],
            [
                [10, 20, 40, 60],
                [40000, 42000, 400, 6000],
                [100, 22000, 4000, 173000],
            ],
        ),
        (
            [{}, {"a": 1}, {"b": 1}, {"a": 1, "b": 1}],
            [
                [random.randint(-128, 127) for _ in range(4)],
                [random.randint(-128, 127) for _ in range(4)],
                [random.randint(-128, 127) for _ in range(4)],
                [random.randint(-32768, 32767) for _ in range(4)],
                [random.randint(-32768, 32767) for _ in range(4)],
                [random.randint(-32768, 32767) for _ in range(4)],
            ],
        ),
    ],
)
def test_onlineVarStoreBuilder(locations, masterValues):
    axisTags = sorted({k for loc in locations for k in loc})
    model = VariationModel(locations)
    builder = OnlineVarStoreBuilder(axisTags)
    builder.setModel(model)
    varIdxs = []
    # shuffle input order to ensure optimizer produces stable results
    random.shuffle(masterValues)
    for masters in masterValues:
        _, varIdx = builder.storeMasters(masters)
        varIdxs.append(varIdx)

    varStore = builder.finish()
    mapping = varStore.optimize()
    varIdxs = [mapping[varIdx] for varIdx in varIdxs]

    dummyFont = TTFont()
    writer = OTTableWriter()
    varStore.compile(writer, dummyFont)
    data = writer.getAllData()
    reader = OTTableReader(data)
    varStore = VarStore()
    varStore.decompile(reader, dummyFont)

    fvarAxes = [buildAxis(axisTag) for axisTag in axisTags]
    instancer = VarStoreInstancer(varStore, fvarAxes)
    for masters, varIdx in zip(masterValues, varIdxs):
        base, *rest = masters
        for expectedValue, loc in zip(masters, locations):
            instancer.setLocation(loc)
            value = base + instancer[varIdx]
            assert expectedValue == value


def buildAxis(axisTag):
    axis = Axis()
    axis.axisTag = axisTag
    return axis


@pytest.mark.parametrize(
    "numRegions, varData, expectedNumVarData, expectedBytes",
    [
        (
            5,
            [
                [10, 10, 0, 0, 20],
                {3: 300},
            ],
            1,
            126,
        ),
        (
            5,
            [
                [10, 10, 0, 0, 20],
                [10, 11, 0, 0, 20],
                [10, 12, 0, 0, 20],
                [10, 13, 0, 0, 20],
                {3: 300},
            ],
            1,
            175,
        ),
        (
            5,
            [
                [10, 11, 0, 0, 20],
                [10, 300, 0, 0, 20],
                [10, 301, 0, 0, 20],
                [10, 302, 0, 0, 20],
                [10, 303, 0, 0, 20],
                [10, 304, 0, 0, 20],
            ],
            1,
            180,
        ),
        (
            5,
            [
                [0, 11, 12, 0, 20],
                [0, 13, 12, 0, 20],
                [0, 14, 12, 0, 20],
                [0, 15, 12, 0, 20],
                [0, 16, 12, 0, 20],
                [10, 300, 0, 0, 20],
                [10, 301, 0, 0, 20],
                [10, 302, 0, 0, 20],
                [10, 303, 0, 0, 20],
                [10, 304, 0, 0, 20],
            ],
            1,
            200,
        ),
        (
            5,
            [
                [0, 11, 12, 0, 20],
                [0, 13, 12, 0, 20],
                [0, 14, 12, 0, 20],
                [0, 15, 12, 0, 20],
                [0, 16, 12, 0, 20],
                [0, 17, 12, 0, 20],
                [0, 18, 12, 0, 20],
                [0, 19, 12, 0, 20],
                [0, 20, 12, 0, 20],
                [10, 300, 0, 0, 20],
                [10, 301, 0, 0, 20],
                [10, 302, 0, 0, 20],
                [10, 303, 0, 0, 20],
                [10, 304, 0, 0, 20],
            ],
            2,
            218,
        ),
        (
            3,
            [
                [10, 10, 10],
            ],
            0,
            12,
        ),
    ],
)
def test_optimize(numRegions, varData, expectedNumVarData, expectedBytes):
    locations = [{i: i / 16384.0} for i in range(numRegions)]
    axisTags = sorted({k for loc in locations for k in loc})

    model = VariationModel(locations)
    builder = OnlineVarStoreBuilder(axisTags)
    builder.setModel(model)

    random.shuffle(varData)
    for data in varData:
        if type(data) is dict:
            newData = [0] * numRegions
            for k, v in data.items():
                newData[k] = v
            data = newData

        builder.storeMasters(data)

    varStore = builder.finish()
    varStore.optimize()

    dummyFont = TTFont()

    writer = XMLWriter(StringIO())
    varStore.toXML(writer, dummyFont)
    xml = writer.file.getvalue()

    assert len(varStore.VarData) == expectedNumVarData, xml

    writer = OTTableWriter()
    varStore.compile(writer, dummyFont)
    data = writer.getAllData()

    assert len(data) == expectedBytes, xml


@pytest.mark.parametrize(
    "quantization, expectedBytes",
    [
        (1, 200),
        (2, 180),
        (3, 170),
        (4, 175),
        (8, 170),
        (32, 92),
        (64, 56),
    ],
)
def test_quantize(quantization, expectedBytes):
    varData = [
        [0, 11, 12, 0, 20],
        [0, 13, 12, 0, 20],
        [0, 14, 12, 0, 20],
        [0, 15, 12, 0, 20],
        [0, 16, 12, 0, 20],
        [10, 300, 0, 0, 20],
        [10, 301, 0, 0, 20],
        [10, 302, 0, 0, 20],
        [10, 303, 0, 0, 20],
        [10, 304, 0, 0, 20],
    ]

    numRegions = 5
    locations = [{i: i / 16384.0} for i in range(numRegions)]
    axisTags = sorted({k for loc in locations for k in loc})

    model = VariationModel(locations)

    builder = OnlineVarStoreBuilder(axisTags)
    builder.setModel(model)

    random.shuffle(varData)
    for data in varData:
        builder.storeMasters(data)

    varStore = builder.finish()
    varStore.optimize(quantization=quantization)

    dummyFont = TTFont()

    writer = XMLWriter(StringIO())
    varStore.toXML(writer, dummyFont)
    xml = writer.file.getvalue()

    writer = OTTableWriter()
    varStore.compile(writer, dummyFont)
    data = writer.getAllData()

    assert len(data) == expectedBytes, xml


def test_optimize_overflow():
    numRegions = 1
    locations = [{"wght": 0}, {"wght": 0.5}]
    axisTags = ["wght"]

    model = VariationModel(locations)
    builder = OnlineVarStoreBuilder(axisTags)
    builder.setModel(model)

    varData = list(range(0, 0xFFFF * 2))
    random.shuffle(varData)
    for data in varData:
        data = [0, data]
        builder.storeMasters(data, round=noRound)

    varStore = builder.finish()
    varStore.optimize()

    for s in varStore.VarData:
        print(len(s.Item))

    # 5 data-sets:
    # - 0..127: 1-byte dataset
    # - 128..32767: 2-byte dataset
    # - 32768..32768+65535-1: 4-byte dataset
    # - 32768+65535..65535+65535-1: 4-byte dataset
    assert len(varStore.VarData) == 4



"""Miscellaneous helpers for our test suite."""

import os
from fontTools.ufoLib.utils import numberTypes


def getDemoFontPath():
    """Return the path to Data/DemoFont.ufo/."""
    testdata = os.path.join(os.path.dirname(__file__), "testdata")
    return os.path.join(testdata, "DemoFont.ufo")


def getDemoFontGlyphSetPath():
    """Return the path to Data/DemoFont.ufo/glyphs/."""
    return os.path.join(getDemoFontPath(), "glyphs")


# GLIF test tools


class Glyph:
    def __init__(self):
        self.name = None
        self.width = None
        self.height = None
        self.unicodes = None
        self.note = None
        self.lib = None
        self.image = None
        self.guidelines = None
        self.anchors = None
        self.outline = []

    def _writePointPenCommand(self, command, args, kwargs):
        args = _listToString(args)
        kwargs = _dictToString(kwargs)
        if args and kwargs:
            return f"pointPen.{command}(*{args}, **{kwargs})"
        elif len(args):
            return f"pointPen.{command}(*{args})"
        elif len(kwargs):
            return f"pointPen.{command}(**{kwargs})"
        else:
            return "pointPen.%s()" % command

    def beginPath(self, **kwargs):
        self.outline.append(self._writePointPenCommand("beginPath", [], kwargs))

    def endPath(self):
        self.outline.append(self._writePointPenCommand("endPath", [], {}))

    def addPoint(self, *args, **kwargs):
        self.outline.append(self._writePointPenCommand("addPoint", args, kwargs))

    def addComponent(self, *args, **kwargs):
        self.outline.append(self._writePointPenCommand("addComponent", args, kwargs))

    def drawPoints(self, pointPen):
        if self.outline:
            py = "\n".join(self.outline)
            exec(py, {"pointPen": pointPen})

    def py(self):
        text = []
        if self.name is not None:
            text.append('glyph.name = "%s"' % self.name)
        if self.width:
            text.append("glyph.width = %r" % self.width)
        if self.height:
            text.append("glyph.height = %r" % self.height)
        if self.unicodes is not None:
            text.append(
                "glyph.unicodes = [%s]" % ", ".join([str(i) for i in self.unicodes])
            )
        if self.note is not None:
            text.append('glyph.note = "%s"' % self.note)
        if self.lib is not None:
            text.append("glyph.lib = %s" % _dictToString(self.lib))
        if self.image is not None:
            text.append("glyph.image = %s" % _dictToString(self.image))
        if self.guidelines is not None:
            text.append("glyph.guidelines = %s" % _listToString(self.guidelines))
        if self.anchors is not None:
            text.append("glyph.anchors = %s" % _listToString(self.anchors))
        if self.outline:
            text += self.outline
        return "\n".join(text)


def _dictToString(d):
    text = []
    for key, value in sorted(d.items()):
        if value is None:
            continue
        key = '"%s"' % key
        if isinstance(value, dict):
            value = _dictToString(value)
        elif isinstance(value, list):
            value = _listToString(value)
        elif isinstance(value, tuple):
            value = _tupleToString(value)
        elif isinstance(value, numberTypes):
            value = repr(value)
        elif isinstance(value, str):
            value = '"%s"' % value
        text.append(f"{key} : {value}")
    if not text:
        return ""
    return "{%s}" % ", ".join(text)


def _listToString(l):
    text = []
    for value in l:
        if isinstance(value, dict):
            value = _dictToString(value)
        elif isinstance(value, list):
            value = _listToString(value)
        elif isinstance(value, tuple):
            value = _tupleToString(value)
        elif isinstance(value, numberTypes):
            value = repr(value)
        elif isinstance(value, str):
            value = '"%s"' % value
        text.append(value)
    if not text:
        return ""
    return "[%s]" % ", ".join(text)


def _tupleToString(t):
    text = []
    for value in t:
        if isinstance(value, dict):
            value = _dictToString(value)
        elif isinstance(value, list):
            value = _listToString(value)
        elif isinstance(value, tuple):
            value = _tupleToString(value)
        elif isinstance(value, numberTypes):
            value = repr(value)
        elif isinstance(value, str):
            value = '"%s"' % value
        text.append(value)
    if not text:
        return ""
    return "(%s)" % ", ".join(text)


def stripText(text):
    new = []
    for line in text.strip().splitlines():
        line = line.strip()
        if not line:
            continue
        new.append(line)
    return "\n".join(new)


# font info values used by several tests

fontInfoVersion1 = {
    "familyName": "Some Font (Family Name)",
    "styleName": "Regular (Style Name)",
    "fullName": "Some Font-Regular (Postscript Full Name)",
    "fontName": "SomeFont-Regular (Postscript Font Name)",
    "menuName": "Some Font Regular (Style Map Family Name)",
    "fontStyle": 64,
    "note": "A note.",
    "versionMajor": 1,
    "versionMinor": 0,
    "year": 2008,
    "copyright": "Copyright Some Foundry.",
    "notice": "Some Font by Some Designer for Some Foundry.",
    "trademark": "Trademark Some Foundry",
    "license": "License info for Some Foundry.",
    "licenseURL": "http://somefoundry.com/license",
    "createdBy": "Some Foundry",
    "designer": "Some Designer",
    "designerURL": "http://somedesigner.com",
    "vendorURL": "http://somefoundry.com",
    "unitsPerEm": 1000,
    "ascender": 750,
    "descender": -250,
    "capHeight": 750,
    "xHeight": 500,
    "defaultWidth": 400,
    "slantAngle": -12.5,
    "italicAngle": -12.5,
    "widthName": "Medium (normal)",
    "weightName": "Medium",
    "weightValue": 500,
    "fondName": "SomeFont Regular (FOND Name)",
    "otFamilyName": "Some Font (Preferred Family Name)",
    "otStyleName": "Regular (Preferred Subfamily Name)",
    "otMacName": "Some Font Regular (Compatible Full Name)",
    "msCharSet": 0,
    "fondID": 15000,
    "uniqueID": 4000000,
    "ttVendor": "SOME",
    "ttUniqueID": "OpenType name Table Unique ID",
    "ttVersion": "OpenType name Table Version",
}

fontInfoVersion2 = {
    "familyName": "Some Font (Family Name)",
    "styleName": "Regular (Style Name)",
    "styleMapFamilyName": "Some Font Regular (Style Map Family Name)",
    "styleMapStyleName": "regular",
    "versionMajor": 1,
    "versionMinor": 0,
    "year": 2008,
    "copyright": "Copyright Some Foundry.",
    "trademark": "Trademark Some Foundry",
    "unitsPerEm": 1000,
    "descender": -250,
    "xHeight": 500,
    "capHeight": 750,
    "ascender": 750,
    "italicAngle": -12.5,
    "note": "A note.",
    "openTypeHeadCreated": "2000/01/01 00:00:00",
    "openTypeHeadLowestRecPPEM": 10,
    "openTypeHeadFlags": [0, 1],
    "openTypeHheaAscender": 750,
    "openTypeHheaDescender": -250,
    "openTypeHheaLineGap": 200,
    "openTypeHheaCaretSlopeRise": 1,
    "openTypeHheaCaretSlopeRun": 0,
    "openTypeHheaCaretOffset": 0,
    "openTypeNameDesigner": "Some Designer",
    "openTypeNameDesignerURL": "http://somedesigner.com",
    "openTypeNameManufacturer": "Some Foundry",
    "openTypeNameManufacturerURL": "http://somefoundry.com",
    "openTypeNameLicense": "License info for Some Foundry.",
    "openTypeNameLicenseURL": "http://somefoundry.com/license",
    "openTypeNameVersion": "OpenType name Table Version",
    "openTypeNameUniqueID": "OpenType name Table Unique ID",
    "openTypeNameDescription": "Some Font by Some Designer for Some Foundry.",
    "openTypeNamePreferredFamilyName": "Some Font (Preferred Family Name)",
    "openTypeNamePreferredSubfamilyName": "Regular (Preferred Subfamily Name)",
    "openTypeNameCompatibleFullName": "Some Font Regular (Compatible Full Name)",
    "openTypeNameSampleText": "Sample Text for Some Font.",
    "openTypeNameWWSFamilyName": "Some Font (WWS Family Name)",
    "openTypeNameWWSSubfamilyName": "Regular (WWS Subfamily Name)",
    "openTypeOS2WidthClass": 5,
    "openTypeOS2WeightClass": 500,
    "openTypeOS2Selection": [3],
    "openTypeOS2VendorID": "SOME",
    "openTypeOS2Panose": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    "openTypeOS2FamilyClass": [1, 1],
    "openTypeOS2UnicodeRanges": [0, 1],
    "openTypeOS2CodePageRanges": [0, 1],
    "openTypeOS2TypoAscender": 750,
    "openTypeOS2TypoDescender": -250,
    "openTypeOS2TypoLineGap": 200,
    "openTypeOS2WinAscent": 750,
    "openTypeOS2WinDescent": 250,
    "openTypeOS2Type": [],
    "openTypeOS2SubscriptXSize": 200,
    "openTypeOS2SubscriptYSize": 400,
    "openTypeOS2SubscriptXOffset": 0,
    "openTypeOS2SubscriptYOffset": -100,
    "openTypeOS2SuperscriptXSize": 200,
    "openTypeOS2SuperscriptYSize": 400,
    "openTypeOS2SuperscriptXOffset": 0,
    "openTypeOS2SuperscriptYOffset": 200,
    "openTypeOS2StrikeoutSize": 20,
    "openTypeOS2StrikeoutPosition": 300,
    "openTypeVheaVertTypoAscender": 750,
    "openTypeVheaVertTypoDescender": -250,
    "openTypeVheaVertTypoLineGap": 200,
    "openTypeVheaCaretSlopeRise": 0,
    "openTypeVheaCaretSlopeRun": 1,
    "openTypeVheaCaretOffset": 0,
    "postscriptFontName": "SomeFont-Regular (Postscript Font Name)",
    "postscriptFullName": "Some Font-Regular (Postscript Full Name)",
    "postscriptSlantAngle": -12.5,
    "postscriptUniqueID": 4000000,
    "postscriptUnderlineThickness": 20,
    "postscriptUnderlinePosition": -200,
    "postscriptIsFixedPitch": False,
    "postscriptBlueValues": [500, 510],
    "postscriptOtherBlues": [-250, -260],
    "postscriptFamilyBlues": [500, 510],
    "postscriptFamilyOtherBlues": [-250, -260],
    "postscriptStemSnapH": [100, 120],
    "postscriptStemSnapV": [80, 90],
    "postscriptBlueFuzz": 1,
    "postscriptBlueShift": 7,
    "postscriptBlueScale": 0.039625,
    "postscriptForceBold": True,
    "postscriptDefaultWidthX": 400,
    "postscriptNominalWidthX": 400,
    "postscriptWeightName": "Medium",
    "postscriptDefaultCharacter": ".notdef",
    "postscriptWindowsCharacterSet": 1,
    "macintoshFONDFamilyID": 15000,
    "macintoshFONDName": "SomeFont Regular (FOND Name)",
}

fontInfoVersion3 = {
    "familyName": "Some Font (Family Name)",
    "styleName": "Regular (Style Name)",
    "styleMapFamilyName": "Some Font Regular (Style Map Family Name)",
    "styleMapStyleName": "regular",
    "versionMajor": 1,
    "versionMinor": 0,
    "year": 2008,
    "copyright": "Copyright Some Foundry.",
    "trademark": "Trademark Some Foundry",
    "unitsPerEm": 1000,
    "descender": -250,
    "xHeight": 500,
    "capHeight": 750,
    "ascender": 750,
    "italicAngle": -12.5,
    "note": "A note.",
    "openTypeGaspRangeRecords": [
        dict(rangeMaxPPEM=10, rangeGaspBehavior=[0]),
        dict(rangeMaxPPEM=20, rangeGaspBehavior=[1]),
        dict(rangeMaxPPEM=30, rangeGaspBehavior=[2]),
        dict(rangeMaxPPEM=40, rangeGaspBehavior=[3]),
        dict(rangeMaxPPEM=50, rangeGaspBehavior=[0, 1, 2, 3]),
        dict(rangeMaxPPEM=0xFFFF, rangeGaspBehavior=[0]),
    ],
    "openTypeHeadCreated": "2000/01/01 00:00:00",
    "openTypeHeadLowestRecPPEM": 10,
    "openTypeHeadFlags": [0, 1],
    "openTypeHheaAscender": 750,
    "openTypeHheaDescender": -250,
    "openTypeHheaLineGap": 200,
    "openTypeHheaCaretSlopeRise": 1,
    "openTypeHheaCaretSlopeRun": 0,
    "openTypeHheaCaretOffset": 0,
    "openTypeNameDesigner": "Some Designer",
    "openTypeNameDesignerURL": "http://somedesigner.com",
    "openTypeNameManufacturer": "Some Foundry",
    "openTypeNameManufacturerURL": "http://somefoundry.com",
    "openTypeNameLicense": "License info for Some Foundry.",
    "openTypeNameLicenseURL": "http://somefoundry.com/license",
    "openTypeNameVersion": "OpenType name Table Version",
    "openTypeNameUniqueID": "OpenType name Table Unique ID",
    "openTypeNameDescription": "Some Font by Some Designer for Some Foundry.",
    "openTypeNamePreferredFamilyName": "Some Font (Preferred Family Name)",
    "openTypeNamePreferredSubfamilyName": "Regular (Preferred Subfamily Name)",
    "openTypeNameCompatibleFullName": "Some Font Regular (Compatible Full Name)",
    "openTypeNameSampleText": "Sample Text for Some Font.",
    "openTypeNameWWSFamilyName": "Some Font (WWS Family Name)",
    "openTypeNameWWSSubfamilyName": "Regular (WWS Subfamily Name)",
    "openTypeNameRecords": [
        dict(nameID=1, platformID=1, encodingID=1, languageID=1, string="Name Record."),
        dict(nameID=2, platformID=1, encodingID=1, languageID=1, string="Name Record."),
    ],
    "openTypeOS2WidthClass": 5,
    "openTypeOS2WeightClass": 500,
    "openTypeOS2Selection": [3],
    "openTypeOS2VendorID": "SOME",
    "openTypeOS2Panose": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    "openTypeOS2FamilyClass": [1, 1],
    "openTypeOS2UnicodeRanges": [0, 1],
    "openTypeOS2CodePageRanges": [0, 1],
    "openTypeOS2TypoAscender": 750,
    "openTypeOS2TypoDescender": -250,
    "openTypeOS2TypoLineGap": 200,
    "openTypeOS2WinAscent": 750,
    "openTypeOS2WinDescent": 250,
    "openTypeOS2Type": [],
    "openTypeOS2SubscriptXSize": 200,
    "openTypeOS2SubscriptYSize": 400,
    "openTypeOS2SubscriptXOffset": 0,
    "openTypeOS2SubscriptYOffset": -100,
    "openTypeOS2SuperscriptXSize": 200,
    "openTypeOS2SuperscriptYSize": 400,
    "openTypeOS2SuperscriptXOffset": 0,
    "openTypeOS2SuperscriptYOffset": 200,
    "openTypeOS2StrikeoutSize": 20,
    "openTypeOS2StrikeoutPosition": 300,
    "openTypeVheaVertTypoAscender": 750,
    "openTypeVheaVertTypoDescender": -250,
    "openTypeVheaVertTypoLineGap": 200,
    "openTypeVheaCaretSlopeRise": 0,
    "openTypeVheaCaretSlopeRun": 1,
    "openTypeVheaCaretOffset": 0,
    "postscriptFontName": "SomeFont-Regular (Postscript Font Name)",
    "postscriptFullName": "Some Font-Regular (Postscript Full Name)",
    "postscriptSlantAngle": -12.5,
    "postscriptUniqueID": 4000000,
    "postscriptUnderlineThickness": 20,
    "postscriptUnderlinePosition": -200,
    "postscriptIsFixedPitch": False,
    "postscriptBlueValues": [500, 510],
    "postscriptOtherBlues": [-250, -260],
    "postscriptFamilyBlues": [500, 510],
    "postscriptFamilyOtherBlues": [-250, -260],
    "postscriptStemSnapH": [100, 120],
    "postscriptStemSnapV": [80, 90],
    "postscriptBlueFuzz": 1,
    "postscriptBlueShift": 7,
    "postscriptBlueScale": 0.039625,
    "postscriptForceBold": True,
    "postscriptDefaultWidthX": 400,
    "postscriptNominalWidthX": 400,
    "postscriptWeightName": "Medium",
    "postscriptDefaultCharacter": ".notdef",
    "postscriptWindowsCharacterSet": 1,
    "macintoshFONDFamilyID": 15000,
    "macintoshFONDName": "SomeFont Regular (FOND Name)",
    "woffMajorVersion": 1,
    "woffMinorVersion": 0,
    "woffMetadataUniqueID": dict(id="string"),
    "woffMetadataVendor": dict(name="Some Foundry", url="http://somefoundry.com"),
    "woffMetadataCredits": dict(
        credits=[
            dict(name="Some Designer"),
            dict(name=""),
            dict(name="Some Designer", url="http://somedesigner.com"),
            dict(name="Some Designer", url=""),
            dict(name="Some Designer", role="Designer"),
            dict(name="Some Designer", role=""),
            dict(name="Some Designer", dir="ltr"),
            dict(name="rengiseD emoS", dir="rtl"),
            {"name": "Some Designer", "class": "hello"},
            {"name": "Some Designer", "class": ""},
        ]
    ),
    "woffMetadataDescription": dict(
        url="http://somefoundry.com/foo/description",
        text=[
            dict(text="foo"),
            dict(text=""),
            dict(text="foo", language="bar"),
            dict(text="foo", language=""),
            dict(text="foo", dir="ltr"),
            dict(text="foo", dir="rtl"),
            {"text": "foo", "class": "foo"},
            {"text": "foo", "class": ""},
        ],
    ),
    "woffMetadataLicense": dict(
        url="http://somefoundry.com/foo/license",
        id="foo",
        text=[
            dict(text="foo"),
            dict(text=""),
            dict(text="foo", language="bar"),
            dict(text="foo", language=""),
            dict(text="foo", dir="ltr"),
            dict(text="foo", dir="rtl"),
            {"text": "foo", "class": "foo"},
            {"text": "foo", "class": ""},
        ],
    ),
    "woffMetadataCopyright": dict(
        text=[
            dict(text="foo"),
            dict(text=""),
            dict(text="foo", language="bar"),
            dict(text="foo", language=""),
            dict(text="foo", dir="ltr"),
            dict(text="foo", dir="rtl"),
            {"text": "foo", "class": "foo"},
            {"text": "foo", "class": ""},
        ]
    ),
    "woffMetadataTrademark": dict(
        text=[
            dict(text="foo"),
            dict(text=""),
            dict(text="foo", language="bar"),
            dict(text="foo", language=""),
            dict(text="foo", dir="ltr"),
            dict(text="foo", dir="rtl"),
            {"text": "foo", "class": "foo"},
            {"text": "foo", "class": ""},
        ]
    ),
    "woffMetadataLicensee": dict(name="Some Licensee"),
    "woffMetadataExtensions": [
        dict(
            # everything
            names=[
                dict(text="foo"),
                dict(text=""),
                dict(text="foo", language="bar"),
                dict(text="foo", language=""),
                dict(text="foo", dir="ltr"),
                dict(text="foo", dir="rtl"),
                {"text": "foo", "class": "hello"},
                {"text": "foo", "class": ""},
            ],
            items=[
                # everything
                dict(
                    id="foo",
                    names=[
                        dict(text="foo"),
                        dict(text=""),
                        dict(text="foo", language="bar"),
                        dict(text="foo", language=""),
                        dict(text="foo", dir="ltr"),
                        dict(text="foo", dir="rtl"),
                        {"text": "foo", "class": "hello"},
                        {"text": "foo", "class": ""},
                    ],
                    values=[
                        dict(text="foo"),
                        dict(text=""),
                        dict(text="foo", language="bar"),
                        dict(text="foo", language=""),
                        dict(text="foo", dir="ltr"),
                        dict(text="foo", dir="rtl"),
                        {"text": "foo", "class": "hello"},
                        {"text": "foo", "class": ""},
                    ],
                ),
                # no id
                dict(names=[dict(text="foo")], values=[dict(text="foo")]),
            ],
        ),
        # no names
        dict(
            items=[dict(id="foo", names=[dict(text="foo")], values=[dict(text="foo")])]
        ),
    ],
    "guidelines": [
        # ints
        dict(x=100, y=200, angle=45),
        # floats
        dict(x=100.5, y=200.5, angle=45.5),
        # edges
        dict(x=0, y=0, angle=0),
        dict(x=0, y=0, angle=360),
        dict(x=0, y=0, angle=360.0),
        # no y
        dict(x=100),
        # no x
        dict(y=200),
        # name
        dict(x=100, y=200, angle=45, name="foo"),
        dict(x=100, y=200, angle=45, name=""),
        # identifier
        dict(x=100, y=200, angle=45, identifier="guide1"),
        dict(x=100, y=200, angle=45, identifier="guide2"),
        dict(x=100, y=200, angle=45, identifier="\x20"),
        dict(x=100, y=200, angle=45, identifier="\x7E"),
        # colors
        dict(x=100, y=200, angle=45, color="0,0,0,0"),
        dict(x=100, y=200, angle=45, color="1,0,0,0"),
        dict(x=100, y=200, angle=45, color="1,1,1,1"),
        dict(x=100, y=200, angle=45, color="0,1,0,0"),
        dict(x=100, y=200, angle=45, color="0,0,1,0"),
        dict(x=100, y=200, angle=45, color="0,0,0,1"),
        dict(x=100, y=200, angle=45, color="1, 0, 0, 0"),
        dict(x=100, y=200, angle=45, color="0, 1, 0, 0"),
        dict(x=100, y=200, angle=45, color="0, 0, 1, 0"),
        dict(x=100, y=200, angle=45, color="0, 0, 0, 1"),
        dict(x=100, y=200, angle=45, color=".5,0,0,0"),
        dict(x=100, y=200, angle=45, color="0,.5,0,0"),
        dict(x=100, y=200, angle=45, color="0,0,.5,0"),
        dict(x=100, y=200, angle=45, color="0,0,0,.5"),
        dict(x=100, y=200, angle=45, color=".5,1,1,1"),
        dict(x=100, y=200, angle=45, color="1,.5,1,1"),
        dict(x=100, y=200, angle=45, color="1,1,.5,1"),
        dict(x=100, y=200, angle=45, color="1,1,1,.5"),
    ],
}

expectedFontInfo1To2Conversion = {
    "familyName": "Some Font (Family Name)",
    "styleMapFamilyName": "Some Font Regular (Style Map Family Name)",
    "styleMapStyleName": "regular",
    "styleName": "Regular (Style Name)",
    "unitsPerEm": 1000,
    "ascender": 750,
    "capHeight": 750,
    "xHeight": 500,
    "descender": -250,
    "italicAngle": -12.5,
    "versionMajor": 1,
    "versionMinor": 0,
    "year": 2008,
    "copyright": "Copyright Some Foundry.",
    "trademark": "Trademark Some Foundry",
    "note": "A note.",
    "macintoshFONDFamilyID": 15000,
    "macintoshFONDName": "SomeFont Regular (FOND Name)",
    "openTypeNameCompatibleFullName": "Some Font Regular (Compatible Full Name)",
    "openTypeNameDescription": "Some Font by Some Designer for Some Foundry.",
    "openTypeNameDesigner": "Some Designer",
    "openTypeNameDesignerURL": "http://somedesigner.com",
    "openTypeNameLicense": "License info for Some Foundry.",
    "openTypeNameLicenseURL": "http://somefoundry.com/license",
    "openTypeNameManufacturer": "Some Foundry",
    "openTypeNameManufacturerURL": "http://somefoundry.com",
    "openTypeNamePreferredFamilyName": "Some Font (Preferred Family Name)",
    "openTypeNamePreferredSubfamilyName": "Regular (Preferred Subfamily Name)",
    "openTypeNameCompatibleFullName": "Some Font Regular (Compatible Full Name)",
    "openTypeNameUniqueID": "OpenType name Table Unique ID",
    "openTypeNameVersion": "OpenType name Table Version",
    "openTypeOS2VendorID": "SOME",
    "openTypeOS2WeightClass": 500,
    "openTypeOS2WidthClass": 5,
    "postscriptDefaultWidthX": 400,
    "postscriptFontName": "SomeFont-Regular (Postscript Font Name)",
    "postscriptFullName": "Some Font-Regular (Postscript Full Name)",
    "postscriptSlantAngle": -12.5,
    "postscriptUniqueID": 4000000,
    "postscriptWeightName": "Medium",
    "postscriptWindowsCharacterSet": 1,
}

expectedFontInfo2To1Conversion = {
    "familyName": "Some Font (Family Name)",
    "menuName": "Some Font Regular (Style Map Family Name)",
    "fontStyle": 64,
    "styleName": "Regular (Style Name)",
    "unitsPerEm": 1000,
    "ascender": 750,
    "capHeight": 750,
    "xHeight": 500,
    "descender": -250,
    "italicAngle": -12.5,
    "versionMajor": 1,
    "versionMinor": 0,
    "copyright": "Copyright Some Foundry.",
    "trademark": "Trademark Some Foundry",
    "note": "A note.",
    "fondID": 15000,
    "fondName": "SomeFont Regular (FOND Name)",
    "fullName": "Some Font Regular (Compatible Full Name)",
    "notice": "Some Font by Some Designer for Some Foundry.",
    "designer": "Some Designer",
    "designerURL": "http://somedesigner.com",
    "license": "License info for Some Foundry.",
    "licenseURL": "http://somefoundry.com/license",
    "createdBy": "Some Foundry",
    "vendorURL": "http://somefoundry.com",
    "otFamilyName": "Some Font (Preferred Family Name)",
    "otStyleName": "Regular (Preferred Subfamily Name)",
    "otMacName": "Some Font Regular (Compatible Full Name)",
    "ttUniqueID": "OpenType name Table Unique ID",
    "ttVersion": "OpenType name Table Version",
    "ttVendor": "SOME",
    "weightValue": 500,
    "widthName": "Medium (normal)",
    "defaultWidth": 400,
    "fontName": "SomeFont-Regular (Postscript Font Name)",
    "fullName": "Some Font-Regular (Postscript Full Name)",
    "slantAngle": -12.5,
    "uniqueID": 4000000,
    "weightName": "Medium",
    "msCharSet": 0,
    "year": 2008,
}



from .interpolatableHelpers import *
from fontTools.ttLib import TTFont
from fontTools.ttLib.ttGlyphSet import LerpGlyphSet
from fontTools.pens.recordingPen import (
    RecordingPen,
    DecomposingRecordingPen,
    RecordingPointPen,
)
from fontTools.pens.boundsPen import ControlBoundsPen
from fontTools.pens.cairoPen import CairoPen
from fontTools.pens.pointPen import (
    SegmentToPointPen,
    PointToSegmentPen,
    ReverseContourPointPen,
)
from itertools import cycle
from io import BytesIO
import cairo
import math
import os
import logging

log = logging.getLogger("fontTools.varLib.interpolatable")


class OverridingDict(dict):
    def __init__(self, parent_dict):
        self.parent_dict = parent_dict

    def __missing__(self, key):
        return self.parent_dict[key]


class InterpolatablePlot:
    width = 8.5 * 72
    height = 11 * 72
    pad = 0.1 * 72
    title_font_size = 24
    font_size = 16
    page_number = 1
    head_color = (0.3, 0.3, 0.3)
    label_color = (0.2, 0.2, 0.2)
    border_color = (0.9, 0.9, 0.9)
    border_width = 0.5
    fill_color = (0.8, 0.8, 0.8)
    stroke_color = (0.1, 0.1, 0.1)
    stroke_width = 1
    oncurve_node_color = (0, 0.8, 0, 0.7)
    oncurve_node_diameter = 6
    offcurve_node_color = (0, 0.5, 0, 0.7)
    offcurve_node_diameter = 4
    handle_color = (0, 0.5, 0, 0.7)
    handle_width = 0.5
    corrected_start_point_color = (0, 0.9, 0, 0.7)
    corrected_start_point_size = 7
    wrong_start_point_color = (1, 0, 0, 0.7)
    start_point_color = (0, 0, 1, 0.7)
    start_arrow_length = 9
    kink_point_size = 7
    kink_point_color = (1, 0, 1, 0.7)
    kink_circle_size = 15
    kink_circle_stroke_width = 1
    kink_circle_color = (1, 0, 1, 0.7)
    contour_colors = ((1, 0, 0), (0, 0, 1), (0, 1, 0), (1, 1, 0), (1, 0, 1), (0, 1, 1))
    contour_alpha = 0.5
    weight_issue_contour_color = (0, 0, 0, 0.4)
    no_issues_label = "Your font's good! Have a cupcake..."
    no_issues_label_color = (0, 0.5, 0)
    cupcake_color = (0.3, 0, 0.3)
    cupcake = r"""
                          ,@.
                        ,@.@@,.
                  ,@@,.@@@.  @.@@@,.
                ,@@. @@@.     @@. @@,.
        ,@@@.@,.@.              @.  @@@@,.@.@@,.
   ,@@.@.     @@.@@.            @,.    .@' @'  @@,
 ,@@. @.          .@@.@@@.  @@'                  @,
,@.  @@.                                          @,
@.     @,@@,.     ,                             .@@,
@,.       .@,@@,.         .@@,.  ,       .@@,  @, @,
@.                             .@. @ @@,.    ,      @
 @,.@@.     @,.      @@,.      @.           @,.    @'
  @@||@,.  @'@,.       @@,.  @@ @,.        @'@@,  @'
     \\@@@@'  @,.      @'@@@@'   @@,.   @@@' //@@@'
      |||||||| @@,.  @@' |||||||  |@@@|@||  ||
       \\\\\\\  ||@@@||  |||||||  |||||||  //
        |||||||  ||||||  ||||||   ||||||  ||
         \\\\\\  ||||||  ||||||  ||||||  //
          ||||||  |||||  |||||   |||||  ||
           \\\\\  |||||  |||||  |||||  //
            |||||  ||||  |||||  ||||  ||
             \\\\  ||||  ||||  ||||  //
              ||||||||||||||||||||||||
"""
    emoticon_color = (0, 0.3, 0.3)
    shrug = r"""\_(")_/"""
    underweight = r"""
 o
/|\
/ \
"""
    overweight = r"""
 o
/O\
/ \
"""
    yay = r""" \o/ """

    def __init__(self, out, glyphsets, names=None, **kwargs):
        self.out = out
        self.glyphsets = glyphsets
        self.names = names or [repr(g) for g in glyphsets]
        self.toc = {}

        for k, v in kwargs.items():
            if not hasattr(self, k):
                raise TypeError("Unknown keyword argument: %s" % k)
            setattr(self, k, v)

        self.panel_width = self.width / 2 - self.pad * 3
        self.panel_height = (
            self.height / 2 - self.pad * 6 - self.font_size * 2 - self.title_font_size
        )

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def show_page(self):
        self.page_number += 1

    def add_title_page(
        self, files, *, show_tolerance=True, tolerance=None, kinkiness=None
    ):
        pad = self.pad
        width = self.width - 3 * self.pad
        x = y = pad

        self.draw_label(
            "Problem report for:",
            x=x,
            y=y,
            bold=True,
            width=width,
            font_size=self.title_font_size,
        )
        y += self.title_font_size

        import hashlib

        for file in files:
            base_file = os.path.basename(file)
            y += self.font_size + self.pad
            self.draw_label(base_file, x=x, y=y, bold=True, width=width)
            y += self.font_size + self.pad

            try:
                with open(file, "rb") as f:
                    h = hashlib.sha1(f.read()).hexdigest()
                self.draw_label("sha1: %s" % h, x=x + pad, y=y, width=width)
                y += self.font_size
            except IsADirectoryError:
                pass

            if file.endswith(".ttf"):
                ttFont = TTFont(file)
                name = ttFont["name"] if "name" in ttFont else None
                if name:
                    for what, nameIDs in (
                        ("Family name", (21, 16, 1)),
                        ("Version", (5,)),
                    ):
                        n = name.getFirstDebugName(nameIDs)
                        if n is None:
                            continue
                        self.draw_label(
                            "%s: %s" % (what, n), x=x + pad, y=y, width=width
                        )
                        y += self.font_size + self.pad
            elif file.endswith((".glyphs", ".glyphspackage")):
                from glyphsLib import GSFont

                f = GSFont(file)
                for what, field in (
                    ("Family name", "familyName"),
                    ("VersionMajor", "versionMajor"),
                    ("VersionMinor", "_versionMinor"),
                ):
                    self.draw_label(
                        "%s: %s" % (what, getattr(f, field)),
                        x=x + pad,
                        y=y,
                        width=width,
                    )
                    y += self.font_size + self.pad

        self.draw_legend(
            show_tolerance=show_tolerance, tolerance=tolerance, kinkiness=kinkiness
        )
        self.show_page()

    def draw_legend(self, *, show_tolerance=True, tolerance=None, kinkiness=None):
        cr = cairo.Context(self.surface)

        x = self.pad
        y = self.height - self.pad - self.font_size * 2
        width = self.width - 2 * self.pad

        xx = x + self.pad * 2
        xxx = x + self.pad * 4

        if show_tolerance:
            self.draw_label(
                "Tolerance: badness; closer to zero the worse", x=xxx, y=y, width=width
            )
            y -= self.pad + self.font_size

        self.draw_label("Underweight contours", x=xxx, y=y, width=width)
        cr.rectangle(xx - self.pad * 0.7, y, 1.5 * self.pad, self.font_size)
        cr.set_source_rgb(*self.fill_color)
        cr.fill_preserve()
        if self.stroke_color:
            cr.set_source_rgb(*self.stroke_color)
            cr.set_line_width(self.stroke_width)
            cr.stroke_preserve()
        cr.set_source_rgba(*self.weight_issue_contour_color)
        cr.fill()
        y -= self.pad + self.font_size

        self.draw_label(
            "Colored contours: contours with the wrong order", x=xxx, y=y, width=width
        )
        cr.rectangle(xx - self.pad * 0.7, y, 1.5 * self.pad, self.font_size)
        if self.fill_color:
            cr.set_source_rgb(*self.fill_color)
            cr.fill_preserve()
        if self.stroke_color:
            cr.set_source_rgb(*self.stroke_color)
            cr.set_line_width(self.stroke_width)
            cr.stroke_preserve()
        cr.set_source_rgba(*self.contour_colors[0], self.contour_alpha)
        cr.fill()
        y -= self.pad + self.font_size

        self.draw_label("Kink artifact", x=xxx, y=y, width=width)
        self.draw_circle(
            cr,
            x=xx,
            y=y + self.font_size * 0.5,
            diameter=self.kink_circle_size,
            stroke_width=self.kink_circle_stroke_width,
            color=self.kink_circle_color,
        )
        y -= self.pad + self.font_size

        self.draw_label("Point causing kink in the contour", x=xxx, y=y, width=width)
        self.draw_dot(
            cr,
            x=xx,
            y=y + self.font_size * 0.5,
            diameter=self.kink_point_size,
            color=self.kink_point_color,
        )
        y -= self.pad + self.font_size

        self.draw_label("Suggested new contour start point", x=xxx, y=y, width=width)
        self.draw_dot(
            cr,
            x=xx,
            y=y + self.font_size * 0.5,
            diameter=self.corrected_start_point_size,
            color=self.corrected_start_point_color,
        )
        y -= self.pad + self.font_size

        self.draw_label(
            "Contour start point in contours with wrong direction",
            x=xxx,
            y=y,
            width=width,
        )
        self.draw_arrow(
            cr,
            x=xx - self.start_arrow_length * 0.3,
            y=y + self.font_size * 0.5,
            color=self.wrong_start_point_color,
        )
        y -= self.pad + self.font_size

        self.draw_label(
            "Contour start point when the first two points overlap",
            x=xxx,
            y=y,
            width=width,
        )
        self.draw_dot(
            cr,
            x=xx,
            y=y + self.font_size * 0.5,
            diameter=self.corrected_start_point_size,
            color=self.start_point_color,
        )
        y -= self.pad + self.font_size

        self.draw_label("Contour start point and direction", x=xxx, y=y, width=width)
        self.draw_arrow(
            cr,
            x=xx - self.start_arrow_length * 0.3,
            y=y + self.font_size * 0.5,
            color=self.start_point_color,
        )
        y -= self.pad + self.font_size

        self.draw_label("Legend:", x=x, y=y, width=width, bold=True)
        y -= self.pad + self.font_size

        if kinkiness is not None:
            self.draw_label(
                "Kink-reporting aggressiveness: %g" % kinkiness,
                x=xxx,
                y=y,
                width=width,
            )
            y -= self.pad + self.font_size

        if tolerance is not None:
            self.draw_label(
                "Error tolerance: %g" % tolerance,
                x=xxx,
                y=y,
                width=width,
            )
            y -= self.pad + self.font_size

        self.draw_label("Parameters:", x=x, y=y, width=width, bold=True)
        y -= self.pad + self.font_size

    def add_summary(self, problems):
        pad = self.pad
        width = self.width - 3 * self.pad
        height = self.height - 2 * self.pad
        x = y = pad

        self.draw_label(
            "Summary of problems",
            x=x,
            y=y,
            bold=True,
            width=width,
            font_size=self.title_font_size,
        )
        y += self.title_font_size

        glyphs_per_problem = defaultdict(set)
        for glyphname, glyph_problems in sorted(problems.items()):
            for problem in glyph_problems:
                glyphs_per_problem[problem["type"]].add(glyphname)

        if InterpolatableProblem.NOTHING in glyphs_per_problem:
            del glyphs_per_problem[InterpolatableProblem.NOTHING]

        for problem_type in sorted(
            glyphs_per_problem, key=lambda x: InterpolatableProblem.severity[x]
        ):
            y += self.font_size
            self.draw_label(
                "%s: %d" % (problem_type, len(glyphs_per_problem[problem_type])),
                x=x,
                y=y,
                width=width,
                bold=True,
            )
            y += self.font_size

            for glyphname in sorted(glyphs_per_problem[problem_type]):
                if y + self.font_size > height:
                    self.show_page()
                    y = self.font_size + pad
                self.draw_label(glyphname, x=x + 2 * pad, y=y, width=width - 2 * pad)
                y += self.font_size

        self.show_page()

    def _add_listing(self, title, items):
        pad = self.pad
        width = self.width - 2 * self.pad
        height = self.height - 2 * self.pad
        x = y = pad

        self.draw_label(
            title, x=x, y=y, bold=True, width=width, font_size=self.title_font_size
        )
        y += self.title_font_size + self.pad

        last_glyphname = None
        for page_no, (glyphname, problems) in items:
            if glyphname == last_glyphname:
                continue
            last_glyphname = glyphname
            if y + self.font_size > height:
                self.show_page()
                y = self.font_size + pad
            self.draw_label(glyphname, x=x + 5 * pad, y=y, width=width - 2 * pad)
            self.draw_label(str(page_no), x=x, y=y, width=4 * pad, align=1)
            y += self.font_size

        self.show_page()

    def add_table_of_contents(self):
        self._add_listing("Table of contents", sorted(self.toc.items()))

    def add_index(self):
        self._add_listing("Index", sorted(self.toc.items(), key=lambda x: x[1][0]))

    def add_problems(self, problems, *, show_tolerance=True, show_page_number=True):
        for glyph, glyph_problems in problems.items():
            last_masters = None
            current_glyph_problems = []
            for p in glyph_problems:
                masters = (
                    p["master_idx"]
                    if "master_idx" in p
                    else (p["master_1_idx"], p["master_2_idx"])
                )
                if masters == last_masters:
                    current_glyph_problems.append(p)
                    continue
                # Flush
                if current_glyph_problems:
                    self.add_problem(
                        glyph,
                        current_glyph_problems,
                        show_tolerance=show_tolerance,
                        show_page_number=show_page_number,
                    )
                    self.show_page()
                    current_glyph_problems = []
                last_masters = masters
                current_glyph_problems.append(p)
            if current_glyph_problems:
                self.add_problem(
                    glyph,
                    current_glyph_problems,
                    show_tolerance=show_tolerance,
                    show_page_number=show_page_number,
                )
                self.show_page()

    def add_problem(
        self, glyphname, problems, *, show_tolerance=True, show_page_number=True
    ):
        if type(problems) not in (list, tuple):
            problems = [problems]

        self.toc[self.page_number] = (glyphname, problems)

        problem_type = problems[0]["type"]
        problem_types = set(problem["type"] for problem in problems)
        if not all(pt == problem_type for pt in problem_types):
            problem_type = ", ".join(sorted({problem["type"] for problem in problems}))

        log.info("Drawing %s: %s", glyphname, problem_type)

        master_keys = (
            ("master_idx",)
            if "master_idx" in problems[0]
            else ("master_1_idx", "master_2_idx")
        )
        master_indices = [problems[0][k] for k in master_keys]

        if problem_type == InterpolatableProblem.MISSING:
            sample_glyph = next(
                i for i, m in enumerate(self.glyphsets) if m[glyphname] is not None
            )
            master_indices.insert(0, sample_glyph)

        x = self.pad
        y = self.pad

        self.draw_label(
            "Glyph name: " + glyphname,
            x=x,
            y=y,
            color=self.head_color,
            align=0,
            bold=True,
            font_size=self.title_font_size,
        )
        tolerance = min(p.get("tolerance", 1) for p in problems)
        if tolerance < 1 and show_tolerance:
            self.draw_label(
                "tolerance: %.2f" % tolerance,
                x=x,
                y=y,
                width=self.width - 2 * self.pad,
                align=1,
                bold=True,
            )
        y += self.title_font_size + self.pad
        self.draw_label(
            "Problems: " + problem_type,
            x=x,
            y=y,
            width=self.width - 2 * self.pad,
            color=self.head_color,
            bold=True,
        )
        y += self.font_size + self.pad * 2

        scales = []
        for which, master_idx in enumerate(master_indices):
            glyphset = self.glyphsets[master_idx]
            name = self.names[master_idx]

            self.draw_label(
                name,
                x=x,
                y=y,
                color=self.label_color,
                width=self.panel_width,
                align=0.5,
            )
            y += self.font_size + self.pad

            if glyphset[glyphname] is not None:
                scales.append(
                    self.draw_glyph(glyphset, glyphname, problems, which, x=x, y=y)
                )
            else:
                self.draw_emoticon(self.shrug, x=x, y=y)
            y += self.panel_height + self.font_size + self.pad

        if any(
            pt
            in (
                InterpolatableProblem.NOTHING,
                InterpolatableProblem.WRONG_START_POINT,
                InterpolatableProblem.CONTOUR_ORDER,
                InterpolatableProblem.KINK,
                InterpolatableProblem.UNDERWEIGHT,
                InterpolatableProblem.OVERWEIGHT,
            )
            for pt in problem_types
        ):
            x = self.pad + self.panel_width + self.pad
            y = self.pad
            y += self.title_font_size + self.pad * 2
            y += self.font_size + self.pad

            glyphset1 = self.glyphsets[master_indices[0]]
            glyphset2 = self.glyphsets[master_indices[1]]

            # Draw the mid-way of the two masters

            self.draw_label(
                "midway interpolation",
                x=x,
                y=y,
                color=self.head_color,
                width=self.panel_width,
                align=0.5,
            )
            y += self.font_size + self.pad

            midway_glyphset = LerpGlyphSet(glyphset1, glyphset2)
            self.draw_glyph(
                midway_glyphset,
                glyphname,
                [{"type": "midway"}]
                + [
                    p
                    for p in problems
                    if p["type"]
                    in (
                        InterpolatableProblem.KINK,
                        InterpolatableProblem.UNDERWEIGHT,
                        InterpolatableProblem.OVERWEIGHT,
                    )
                ],
                None,
                x=x,
                y=y,
                scale=min(scales),
            )

            y += self.panel_height + self.font_size + self.pad

        if any(
            pt
            in (
                InterpolatableProblem.WRONG_START_POINT,
                InterpolatableProblem.CONTOUR_ORDER,
                InterpolatableProblem.KINK,
            )
            for pt in problem_types
        ):
            # Draw the proposed fix

            self.draw_label(
                "proposed fix",
                x=x,
                y=y,
                color=self.head_color,
                width=self.panel_width,
                align=0.5,
            )
            y += self.font_size + self.pad

            overriding1 = OverridingDict(glyphset1)
            overriding2 = OverridingDict(glyphset2)
            perContourPen1 = PerContourOrComponentPen(
                RecordingPen, glyphset=overriding1
            )
            perContourPen2 = PerContourOrComponentPen(
                RecordingPen, glyphset=overriding2
            )
            glyphset1[glyphname].draw(perContourPen1)
            glyphset2[glyphname].draw(perContourPen2)

            for problem in problems:
                if problem["type"] == InterpolatableProblem.CONTOUR_ORDER:
                    fixed_contours = [
                        perContourPen2.value[i] for i in problems[0]["value_2"]
                    ]
                    perContourPen2.value = fixed_contours

            for problem in problems:
                if problem["type"] == InterpolatableProblem.WRONG_START_POINT:
                    # Save the wrong contours
                    wrongContour1 = perContourPen1.value[problem["contour"]]
                    wrongContour2 = perContourPen2.value[problem["contour"]]

                    # Convert the wrong contours to point pens
                    points1 = RecordingPointPen()
                    converter = SegmentToPointPen(points1, False)
                    wrongContour1.replay(converter)
                    points2 = RecordingPointPen()
                    converter = SegmentToPointPen(points2, False)
                    wrongContour2.replay(converter)

                    proposed_start = problem["value_2"]

                    # See if we need reversing; fragile but worth a try
                    if problem["reversed"]:
                        new_points2 = RecordingPointPen()
                        reversedPen = ReverseContourPointPen(new_points2)
                        points2.replay(reversedPen)
                        points2 = new_points2
                        proposed_start = len(points2.value) - 2 - proposed_start

                    # Rotate points2 so that the first point is the same as in points1
                    beginPath = points2.value[:1]
                    endPath = points2.value[-1:]
                    pts = points2.value[1:-1]
                    pts = pts[proposed_start:] + pts[:proposed_start]
                    points2.value = beginPath + pts + endPath

                    # Convert the point pens back to segment pens
                    segment1 = RecordingPen()
                    converter = PointToSegmentPen(segment1, True)
                    points1.replay(converter)
                    segment2 = RecordingPen()
                    converter = PointToSegmentPen(segment2, True)
                    points2.replay(converter)

                    # Replace the wrong contours
                    wrongContour1.value = segment1.value
                    wrongContour2.value = segment2.value
                    perContourPen1.value[problem["contour"]] = wrongContour1
                    perContourPen2.value[problem["contour"]] = wrongContour2

            for problem in problems:
                # If we have a kink, try to fix it.
                if problem["type"] == InterpolatableProblem.KINK:
                    # Save the wrong contours
                    wrongContour1 = perContourPen1.value[problem["contour"]]
                    wrongContour2 = perContourPen2.value[problem["contour"]]

                    # Convert the wrong contours to point pens
                    points1 = RecordingPointPen()
                    converter = SegmentToPointPen(points1, False)
                    wrongContour1.replay(converter)
                    points2 = RecordingPointPen()
                    converter = SegmentToPointPen(points2, False)
                    wrongContour2.replay(converter)

                    i = problem["value"]

                    # Position points to be around the same ratio
                    # beginPath / endPath dance
                    j = i + 1
                    pt0 = points1.value[j][1][0]
                    pt1 = points2.value[j][1][0]
                    j_prev = (i - 1) % (len(points1.value) - 2) + 1
                    pt0_prev = points1.value[j_prev][1][0]
                    pt1_prev = points2.value[j_prev][1][0]
                    j_next = (i + 1) % (len(points1.value) - 2) + 1
                    pt0_next = points1.value[j_next][1][0]
                    pt1_next = points2.value[j_next][1][0]

                    pt0 = complex(*pt0)
                    pt1 = complex(*pt1)
                    pt0_prev = complex(*pt0_prev)
                    pt1_prev = complex(*pt1_prev)
                    pt0_next = complex(*pt0_next)
                    pt1_next = complex(*pt1_next)

                    # Find the ratio of the distance between the points
                    r0 = abs(pt0 - pt0_prev) / abs(pt0_next - pt0_prev)
                    r1 = abs(pt1 - pt1_prev) / abs(pt1_next - pt1_prev)
                    r_mid = (r0 + r1) / 2

                    pt0 = pt0_prev + r_mid * (pt0_next - pt0_prev)
                    pt1 = pt1_prev + r_mid * (pt1_next - pt1_prev)

                    points1.value[j] = (
                        points1.value[j][0],
                        (((pt0.real, pt0.imag),) + points1.value[j][1][1:]),
                        points1.value[j][2],
                    )
                    points2.value[j] = (
                        points2.value[j][0],
                        (((pt1.real, pt1.imag),) + points2.value[j][1][1:]),
                        points2.value[j][2],
                    )

                    # Convert the point pens back to segment pens
                    segment1 = RecordingPen()
                    converter = PointToSegmentPen(segment1, True)
                    points1.replay(converter)
                    segment2 = RecordingPen()
                    converter = PointToSegmentPen(segment2, True)
                    points2.replay(converter)

                    # Replace the wrong contours
                    wrongContour1.value = segment1.value
                    wrongContour2.value = segment2.value

            # Assemble
            fixed1 = RecordingPen()
            fixed2 = RecordingPen()
            for contour in perContourPen1.value:
                fixed1.value.extend(contour.value)
            for contour in perContourPen2.value:
                fixed2.value.extend(contour.value)
            fixed1.draw = fixed1.replay
            fixed2.draw = fixed2.replay

            overriding1[glyphname] = fixed1
            overriding2[glyphname] = fixed2

            try:
                midway_glyphset = LerpGlyphSet(overriding1, overriding2)
                self.draw_glyph(
                    midway_glyphset,
                    glyphname,
                    {"type": "fixed"},
                    None,
                    x=x,
                    y=y,
                    scale=min(scales),
                )
            except ValueError:
                self.draw_emoticon(self.shrug, x=x, y=y)
            y += self.panel_height + self.pad

        else:
            emoticon = self.shrug
            if InterpolatableProblem.UNDERWEIGHT in problem_types:
                emoticon = self.underweight
            elif InterpolatableProblem.OVERWEIGHT in problem_types:
                emoticon = self.overweight
            elif InterpolatableProblem.NOTHING in problem_types:
                emoticon = self.yay
            self.draw_emoticon(emoticon, x=x, y=y)

        if show_page_number:
            self.draw_label(
                str(self.page_number),
                x=0,
                y=self.height - self.font_size - self.pad,
                width=self.width,
                color=self.head_color,
                align=0.5,
            )

    def draw_label(
        self,
        label,
        *,
        x=0,
        y=0,
        color=(0, 0, 0),
        align=0,
        bold=False,
        width=None,
        height=None,
        font_size=None,
    ):
        if width is None:
            width = self.width
        if height is None:
            height = self.height
        if font_size is None:
            font_size = self.font_size
        cr = cairo.Context(self.surface)
        cr.select_font_face(
            "@cairo:",
            cairo.FONT_SLANT_NORMAL,
            cairo.FONT_WEIGHT_BOLD if bold else cairo.FONT_WEIGHT_NORMAL,
        )
        cr.set_font_size(font_size)
        font_extents = cr.font_extents()
        font_size = font_size * font_size / font_extents[2]
        cr.set_font_size(font_size)
        font_extents = cr.font_extents()

        cr.set_source_rgb(*color)

        extents = cr.text_extents(label)
        if extents.width > width:
            # Shrink
            font_size *= width / extents.width
            cr.set_font_size(font_size)
            font_extents = cr.font_extents()
            extents = cr.text_extents(label)

        # Center
        label_x = x + (width - extents.width) * align
        label_y = y + font_extents[0]
        cr.move_to(label_x, label_y)
        cr.show_text(label)

    def draw_glyph(self, glyphset, glyphname, problems, which, *, x=0, y=0, scale=None):
        if type(problems) not in (list, tuple):
            problems = [problems]

        midway = any(problem["type"] == "midway" for problem in problems)
        problem_type = problems[0]["type"]
        problem_types = set(problem["type"] for problem in problems)
        if not all(pt == problem_type for pt in problem_types):
            problem_type = "mixed"
        glyph = glyphset[glyphname]

        recording = RecordingPen()
        glyph.draw(recording)
        decomposedRecording = DecomposingRecordingPen(glyphset)
        glyph.draw(decomposedRecording)

        boundsPen = ControlBoundsPen(glyphset)
        decomposedRecording.replay(boundsPen)
        bounds = boundsPen.bounds
        if bounds is None:
            bounds = (0, 0, 0, 0)

        glyph_width = bounds[2] - bounds[0]
        glyph_height = bounds[3] - bounds[1]

        if glyph_width:
            if scale is None:
                scale = self.panel_width / glyph_width
            else:
                scale = min(scale, self.panel_width / glyph_width)
        if glyph_height:
            if scale is None:
                scale = self.panel_height / glyph_height
            else:
                scale = min(scale, self.panel_height / glyph_height)
        if scale is None:
            scale = 1

        cr = cairo.Context(self.surface)
        cr.translate(x, y)
        # Center
        cr.translate(
            (self.panel_width - glyph_width * scale) / 2,
            (self.panel_height - glyph_height * scale) / 2,
        )
        cr.scale(scale, -scale)
        cr.translate(-bounds[0], -bounds[3])

        if self.border_color:
            cr.set_source_rgb(*self.border_color)
            cr.rectangle(bounds[0], bounds[1], glyph_width, glyph_height)
            cr.set_line_width(self.border_width / scale)
            cr.stroke()

        if self.fill_color or self.stroke_color:
            pen = CairoPen(glyphset, cr)
            decomposedRecording.replay(pen)

            if self.fill_color and problem_type != InterpolatableProblem.OPEN_PATH:
                cr.set_source_rgb(*self.fill_color)
                cr.fill_preserve()

            if self.stroke_color:
                cr.set_source_rgb(*self.stroke_color)
                cr.set_line_width(self.stroke_width / scale)
                cr.stroke_preserve()

            cr.new_path()

        if (
            InterpolatableProblem.UNDERWEIGHT in problem_types
            or InterpolatableProblem.OVERWEIGHT in problem_types
        ):
            perContourPen = PerContourOrComponentPen(RecordingPen, glyphset=glyphset)
            recording.replay(perContourPen)
            for problem in problems:
                if problem["type"] in (
                    InterpolatableProblem.UNDERWEIGHT,
                    InterpolatableProblem.OVERWEIGHT,
                ):
                    contour = perContourPen.value[problem["contour"]]
                    contour.replay(CairoPen(glyphset, cr))
                    cr.set_source_rgba(*self.weight_issue_contour_color)
                    cr.fill()

        if any(
            t in problem_types
            for t in {
                InterpolatableProblem.NOTHING,
                InterpolatableProblem.NODE_COUNT,
                InterpolatableProblem.NODE_INCOMPATIBILITY,
            }
        ):
            cr.set_line_cap(cairo.LINE_CAP_ROUND)

            # Oncurve nodes
            for segment, args in decomposedRecording.value:
                if not args:
                    continue
                x, y = args[-1]
                cr.move_to(x, y)
                cr.line_to(x, y)
            cr.set_source_rgba(*self.oncurve_node_color)
            cr.set_line_width(self.oncurve_node_diameter / scale)
            cr.stroke()

            # Offcurve nodes
            for segment, args in decomposedRecording.value:
                if not args:
                    continue
                for x, y in args[:-1]:
                    cr.move_to(x, y)
                    cr.line_to(x, y)
            cr.set_source_rgba(*self.offcurve_node_color)
            cr.set_line_width(self.offcurve_node_diameter / scale)
            cr.stroke()

            # Handles
            for segment, args in decomposedRecording.value:
                if not args:
                    pass
                elif segment in ("moveTo", "lineTo"):
                    cr.move_to(*args[0])
                elif segment == "qCurveTo":
                    for x, y in args:
                        cr.line_to(x, y)
                    cr.new_sub_path()
                    cr.move_to(*args[-1])
                elif segment == "curveTo":
                    cr.line_to(*args[0])
                    cr.new_sub_path()
                    cr.move_to(*args[1])
                    cr.line_to(*args[2])
                    cr.new_sub_path()
                    cr.move_to(*args[-1])
                else:
                    continue

            cr.set_source_rgba(*self.handle_color)
            cr.set_line_width(self.handle_width / scale)
            cr.stroke()

        matching = None
        for problem in problems:
            if problem["type"] == InterpolatableProblem.CONTOUR_ORDER:
                matching = problem["value_2"]
                colors = cycle(self.contour_colors)
                perContourPen = PerContourOrComponentPen(
                    RecordingPen, glyphset=glyphset
                )
                recording.replay(perContourPen)
                for i, contour in enumerate(perContourPen.value):
                    if matching[i] == i:
                        continue
                    color = next(colors)
                    contour.replay(CairoPen(glyphset, cr))
                    cr.set_source_rgba(*color, self.contour_alpha)
                    cr.fill()

        for problem in problems:
            if problem["type"] in (
                InterpolatableProblem.NOTHING,
                InterpolatableProblem.WRONG_START_POINT,
            ):
                idx = problem.get("contour")

                # Draw suggested point
                if idx is not None and which == 1 and "value_2" in problem:
                    perContourPen = PerContourOrComponentPen(
                        RecordingPen, glyphset=glyphset
                    )
                    decomposedRecording.replay(perContourPen)
                    points = SimpleRecordingPointPen()
                    converter = SegmentToPointPen(points, False)
                    perContourPen.value[
                        idx if matching is None else matching[idx]
                    ].replay(converter)
                    targetPoint = points.value[problem["value_2"]][0]
                    cr.save()
                    cr.translate(*targetPoint)
                    cr.scale(1 / scale, 1 / scale)
                    self.draw_dot(
                        cr,
                        diameter=self.corrected_start_point_size,
                        color=self.corrected_start_point_color,
                    )
                    cr.restore()

                # Draw start-point arrow
                if which == 0 or not problem.get("reversed"):
                    color = self.start_point_color
                else:
                    color = self.wrong_start_point_color
                first_pt = None
                i = 0
                cr.save()
                for segment, args in decomposedRecording.value:
                    if segment == "moveTo":
                        first_pt = args[0]
                        continue
                    if first_pt is None:
                        continue
                    if segment == "closePath":
                        second_pt = first_pt
                    else:
                        second_pt = args[0]

                    if idx is None or i == idx:
                        cr.save()
                        first_pt = complex(*first_pt)
                        second_pt = complex(*second_pt)
                        length = abs(second_pt - first_pt)
                        cr.translate(first_pt.real, first_pt.imag)
                        if length:
                            # Draw arrowhead
                            cr.rotate(
                                math.atan2(
                                    second_pt.imag - first_pt.imag,
                                    second_pt.real - first_pt.real,
                                )
                            )
                            cr.scale(1 / scale, 1 / scale)
                            self.draw_arrow(cr, color=color)
                        else:
                            # Draw circle
                            cr.scale(1 / scale, 1 / scale)
                            self.draw_dot(
                                cr,
                                diameter=self.corrected_start_point_size,
                                color=color,
                            )
                        cr.restore()

                        if idx is not None:
                            break

                    first_pt = None
                    i += 1

                cr.restore()

            if problem["type"] == InterpolatableProblem.KINK:
                idx = problem.get("contour")
                perContourPen = PerContourOrComponentPen(
                    RecordingPen, glyphset=glyphset
                )
                decomposedRecording.replay(perContourPen)
                points = SimpleRecordingPointPen()
                converter = SegmentToPointPen(points, False)
                perContourPen.value[idx if matching is None else matching[idx]].replay(
                    converter
                )

                targetPoint = points.value[problem["value"]][0]
                cr.save()
                cr.translate(*targetPoint)
                cr.scale(1 / scale, 1 / scale)
                if midway:
                    self.draw_circle(
                        cr,
                        diameter=self.kink_circle_size,
                        stroke_width=self.kink_circle_stroke_width,
                        color=self.kink_circle_color,
                    )
                else:
                    self.draw_dot(
                        cr,
                        diameter=self.kink_point_size,
                        color=self.kink_point_color,
                    )
                cr.restore()

        return scale

    def draw_dot(self, cr, *, x=0, y=0, color=(0, 0, 0), diameter=10):
        cr.save()
        cr.set_line_width(diameter)
        cr.set_line_cap(cairo.LINE_CAP_ROUND)
        cr.move_to(x, y)
        cr.line_to(x, y)
        if len(color) == 3:
            color = color + (1,)
        cr.set_source_rgba(*color)
        cr.stroke()
        cr.restore()

    def draw_circle(
        self, cr, *, x=0, y=0, color=(0, 0, 0), diameter=10, stroke_width=1
    ):
        cr.save()
        cr.set_line_width(stroke_width)
        cr.set_line_cap(cairo.LINE_CAP_SQUARE)
        cr.arc(x, y, diameter / 2, 0, 2 * math.pi)
        if len(color) == 3:
            color = color + (1,)
        cr.set_source_rgba(*color)
        cr.stroke()
        cr.restore()

    def draw_arrow(self, cr, *, x=0, y=0, color=(0, 0, 0)):
        cr.save()
        if len(color) == 3:
            color = color + (1,)
        cr.set_source_rgba(*color)
        cr.translate(self.start_arrow_length + x, y)
        cr.move_to(0, 0)
        cr.line_to(
            -self.start_arrow_length,
            -self.start_arrow_length * 0.4,
        )
        cr.line_to(
            -self.start_arrow_length,
            self.start_arrow_length * 0.4,
        )
        cr.close_path()
        cr.fill()
        cr.restore()

    def draw_text(self, text, *, x=0, y=0, color=(0, 0, 0), width=None, height=None):
        if width is None:
            width = self.width
        if height is None:
            height = self.height

        text = text.splitlines()
        cr = cairo.Context(self.surface)
        cr.set_source_rgb(*color)
        cr.set_font_size(self.font_size)
        cr.select_font_face(
            "@cairo:monospace", cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_NORMAL
        )
        text_width = 0
        text_height = 0
        font_extents = cr.font_extents()
        font_font_size = font_extents[2]
        font_ascent = font_extents[0]
        for line in text:
            extents = cr.text_extents(line)
            text_width = max(text_width, extents.x_advance)
            text_height += font_font_size
        if not text_width:
            return
        cr.translate(x, y)
        scale = min(width / text_width, height / text_height)
        # center
        cr.translate(
            (width - text_width * scale) / 2, (height - text_height * scale) / 2
        )
        cr.scale(scale, scale)

        cr.translate(0, font_ascent)
        for line in text:
            cr.move_to(0, 0)
            cr.show_text(line)
            cr.translate(0, font_font_size)

    def draw_cupcake(self):
        self.draw_label(
            self.no_issues_label,
            x=self.pad,
            y=self.pad,
            color=self.no_issues_label_color,
            width=self.width - 2 * self.pad,
            align=0.5,
            bold=True,
            font_size=self.title_font_size,
        )

        self.draw_text(
            self.cupcake,
            x=self.pad,
            y=self.pad + self.font_size,
            width=self.width - 2 * self.pad,
            height=self.height - 2 * self.pad - self.font_size,
            color=self.cupcake_color,
        )

    def draw_emoticon(self, emoticon, x=0, y=0):
        self.draw_text(
            emoticon,
            x=x,
            y=y,
            color=self.emoticon_color,
            width=self.panel_width,
            height=self.panel_height,
        )


class InterpolatablePostscriptLike(InterpolatablePlot):
    def __exit__(self, type, value, traceback):
        self.surface.finish()

    def show_page(self):
        super().show_page()
        self.surface.show_page()


class InterpolatablePS(InterpolatablePostscriptLike):
    def __enter__(self):
        self.surface = cairo.PSSurface(self.out, self.width, self.height)
        return self


class InterpolatablePDF(InterpolatablePostscriptLike):
    def __enter__(self):
        self.surface = cairo.PDFSurface(self.out, self.width, self.height)
        self.surface.set_metadata(
            cairo.PDF_METADATA_CREATOR, "fonttools varLib.interpolatable"
        )
        self.surface.set_metadata(cairo.PDF_METADATA_CREATE_DATE, "")
        return self


class InterpolatableSVG(InterpolatablePlot):
    def __enter__(self):
        self.sink = BytesIO()
        self.surface = cairo.SVGSurface(self.sink, self.width, self.height)
        return self

    def __exit__(self, type, value, traceback):
        if self.surface is not None:
            self.show_page()

    def show_page(self):
        super().show_page()
        self.surface.finish()
        self.out.append(self.sink.getvalue())
        self.sink = BytesIO()
        self.surface = cairo.SVGSurface(self.sink, self.width, self.height)



from fontTools.misc.textTools import safeEval
from . import DefaultTable
import struct


GASP_SYMMETRIC_GRIDFIT = 0x0004
GASP_SYMMETRIC_SMOOTHING = 0x0008
GASP_DOGRAY = 0x0002
GASP_GRIDFIT = 0x0001


class table__g_a_s_p(DefaultTable.DefaultTable):
    """Grid-fitting and Scan-conversion Procedure table

    The ``gasp`` table defines the preferred rasterization settings for
    the font when rendered on monochrome and greyscale output devices.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/gasp
    """

    def decompile(self, data, ttFont):
        self.version, numRanges = struct.unpack(">HH", data[:4])
        assert 0 <= self.version <= 1, "unknown 'gasp' format: %s" % self.version
        data = data[4:]
        self.gaspRange = {}
        for i in range(numRanges):
            rangeMaxPPEM, rangeGaspBehavior = struct.unpack(">HH", data[:4])
            self.gaspRange[int(rangeMaxPPEM)] = int(rangeGaspBehavior)
            data = data[4:]
        assert not data, "too much data"

    def compile(self, ttFont):
        version = 0  # ignore self.version
        numRanges = len(self.gaspRange)
        data = b""
        items = sorted(self.gaspRange.items())
        for rangeMaxPPEM, rangeGaspBehavior in items:
            data = data + struct.pack(">HH", rangeMaxPPEM, rangeGaspBehavior)
            if rangeGaspBehavior & ~(GASP_GRIDFIT | GASP_DOGRAY):
                version = 1
        data = struct.pack(">HH", version, numRanges) + data
        return data

    def toXML(self, writer, ttFont):
        items = sorted(self.gaspRange.items())
        for rangeMaxPPEM, rangeGaspBehavior in items:
            writer.simpletag(
                "gaspRange",
                [
                    ("rangeMaxPPEM", rangeMaxPPEM),
                    ("rangeGaspBehavior", rangeGaspBehavior),
                ],
            )
            writer.newline()

    def fromXML(self, name, attrs, content, ttFont):
        if name != "gaspRange":
            return
        if not hasattr(self, "gaspRange"):
            self.gaspRange = {}
        self.gaspRange[safeEval(attrs["rangeMaxPPEM"])] = safeEval(
            attrs["rangeGaspBehavior"]
        )



from fontTools.ttLib.tables._c_v_t import table__c_v_t


def test_compile_without_values():
    # When read from XML, in an empty cvt table, the values attribute is missing.
    # Make sure that the table can be compiled to a zero-length table.
    cvt = table__c_v_t()
    assert cvt.compile(None) == b""


def test_decompile_without_values():
    # VTT adds a zero-length table to a font if the Control Values are empty.
    # Make sure it can be decompiled.
    cvt = table__c_v_t()
    cvt.decompile(b"", None)



#! /usr/bin/env python3

# Sample script to convert legacy cmap subtables to format-4
# subtables.  Note that this is rarely what one needs.  You
# probably need to just drop the legacy subtables if the font
# already has a format-4 subtable.
#
# Other times, you would need to convert a non-Unicode cmap
# legacy subtable to a Unicode one.  In those cases, use the
# getEncoding() of subtable and use that encoding to map the
# characters to Unicode...  TODO: Extend this script to do that.

from fontTools.ttLib import TTFont
from fontTools.ttLib.tables._c_m_a_p import CmapSubtable
import sys

if len(sys.argv) != 3:
    print("usage: cmap-format.py fontfile.ttf outfile.ttf")
    sys.exit(1)
fontfile = sys.argv[1]
outfile = sys.argv[2]
font = TTFont(fontfile)

cmap = font["cmap"]
outtables = []
for table in cmap.tables:
    if table.format in [4, 12, 13, 14]:
        outtables.append(table)
    # Convert ot format4
    newtable = CmapSubtable.newSubtable(4)
    newtable.platformID = table.platformID
    newtable.platEncID = table.platEncID
    newtable.language = table.language
    newtable.cmap = table.cmap
    outtables.append(newtable)
cmap.tables = outtables

font.save(outfile)
